from .default_embedders import (
                                ContinuousEmbedder,
                                IntegerEmbedder,
                                CategoricalEmbedder,
                                IdentityEmbedder
                                )




import numpy as np
from parmoo import MOOP
from parmoo.acquisitions import RandomConstraint, FixedWeights
from parmoo.searches import LatinHypercube
from parmoo.surrogates import GaussRBF
from parmoo.optimizers import GlobalSurrogate_BFGS

# Create a new MOOP with a derivative-based solver
my_moop = MOOP(GlobalSurrogate_BFGS,
               # Use the hyperparams to fix the random seed for reproducibility
               hyperparams={'np_random_gen': 0})

# Add 3 continuous variables named x1, x2, x3
for i in range(3):
    my_moop.addDesign({'name': f"x{i+1}",
                       'des_type': "continuous",
                       'lb': 0.0,
                       'ub': 1.0,
                       'des_tol': 1.0e-8})
# Add one categorical variable named x4
my_moop.addDesign({'name': "x4",
                   'des_type': "categorical",
                   'levels': 3})

def quad_sim(x):
    """ A quadratic simulation function with 2 outputs.

    Returns:
        np.ndarray: simulation value (S) with 2 outputs
         * S_1(x) = <x, x>
         * S_2(x) = <x-1, x-1>

    """

    return np.array([x["x1"] ** 2 + x["x2"] ** 2 + x["x3"] ** 2,
                     (x["x1"] - 1.0) ** 2 + (x["x2"] - 1.0) ** 2 +
                     (x["x3"] - 1.0) ** 2])

# Add the quadratic simulation to the problem
# Use a 10-point LHS experimental design and a Gaussian RBF surrogate model
my_moop.addSimulation({'name': "f_conv",
                       'm': 2,
                       'sim_func': quad_sim,
                       'search': LatinHypercube,
                       'surrogate': GaussRBF,
                       'hyperparams': {'search_budget': 10}})

# Define some objectives below -- try to avoid things that jax can't compile

def obj_f1_func(x, sim):
    """ Minimize the first output from 'f_conv' """

    return sim['f_conv'][0]

def obj_f1_grad(x, sim):
    """ Corresponding gradient evaluations for obj_f1_func """

    dx = {'x1': 0.0, 'x2': 0.0, 'x3': 0.0, 'x4': 0.0}
    ds = {'f_conv': np.eye(2)[0]}
    return dx, ds

def obj_f2_func(x, sim):
    """ Minimize the second output from 'f_conv' """

    return sim['f_conv'][1]

def obj_f2_grad(x, sim):
    """ Corresponding gradient evaluations for obj_f2_func """

    dx = {'x1': 0.0, 'x2': 0.0, 'x3': 0.0, 'x4': 0.0}
    ds = {'f_conv': np.eye(2)[1]}
    return dx, ds

# Minimize each of the 2 outputs from the quadratic simulation
my_moop.addObjective({'name': "f1",
                      'obj_func': obj_f1_func,
                      'obj_grad': obj_f1_grad})
my_moop.addObjective({'name': "f2",
                      'obj_func': obj_f2_func,
                      'obj_grad': obj_f2_grad})

def const_x4_func(x, sim):
    """ Constrain x["x4"] = 0 """

    return 1.0 - (x["x4"] == 0)

def const_x4_grad(x, sim):
    """ Gradient for evaluating whether x["x4"] = 0 """

    # Note: There is no partial derivative for a categorical design variable.
    # This may make it hard to solve problems that place constraints on many
    # categorical variables, but for 1 categorical variable it should be okay.
    # We can just set all gradients equal to 0.
    dx = {'x1': 0.0, 'x2': 0.0, 'x3': 0.0, 'x4': 0.0}
    ds = {'f_conv': np.zeros(2)}
    return dx, ds

# Add the single constraint to the problem
my_moop.addConstraint({'name': "c_x4",
                       'con_func': const_x4_func,
                       'con_grad': const_x4_grad})

# Add 2 different acquisition functions to the problem
my_moop.addAcquisition({'acquisition': RandomConstraint})
my_moop.addAcquisition({'acquisition': FixedWeights,
                        # Fixed weights with equal weight on both objectives
                        'hyperparams': {'weights': np.array([0.5, 0.5])}})

# Turn on checkpointing -- creates the files parmoo.moop and parmoo.surrogate.1
my_moop.setCheckpoint(True, checkpoint_data=False, filename="parmoo")

# Turn on logging
import logging
logging.basicConfig(level=logging.INFO,
                    format='%(asctime)s %(levelname)-8s %(message)s',
                    datefmt='%Y-%m-%d %H:%M:%S')

# Solve the problem
my_moop.solve(5)

# Get and print full simulation database
sim_db = my_moop.getSimulationData(format="pandas")
print("Simulation data:")
for key in sim_db.keys():
    print(f"\t{key}:")
    print(sim_db[key])

# Get and print results
soln = my_moop.getPF(format="pandas")
print("\n\n")
print("Solution points:")
print(soln)




def test_MOOP_embed_extract():
    """ Test that the MOOP class can embed/extract design variables.

    Add several design variables and generate an embedding. Then embed and
    extract several inputs, and check that the results match up to the
    design tolerance.

    """

    from jax import config
    config.update("jax_enable_x64", True)
    import numpy as np
    from parmoo import MOOP
    from parmoo.embeddings import IdentityEmbedder
    from parmoo.optimizers import LocalSurrogate_PS

    # Create a MOOP with 6 variables of mixed types
    moop = MOOP(LocalSurrogate_PS)
    moop.addDesign({'des_type': "integer",
                    'lb': 0,
                    'ub': 1000})
    moop.addDesign({'des_type': "continuous",
                    'lb': -1.0,
                    'ub': 0.0})
    moop.addDesign({'des_type': "categorical",
                    'levels': 2})
    moop.addDesign({'des_type': "categorical",
                    'levels': ["0guy", "1guy", "2guy"]})
    moop.addDesign({'des_type': "custom",
                    'embedder': IdentityEmbedder,
                    'lb': 0.0, 'ub': 1.0})
    moop.addDesign({'des_type': "raw",
                    'lb': 0.0, 'ub': 5.0})
    moop.addDesign({'des_type': "continuous",
                    'lb': 0.0,
                    'ub': 1.0})
    # Embed/extract lower bounds, upper bounds, and 5 random values
    test_pts = np.zeros((7, 7))
    test_pts[1, :] = 1.0
    test_pts[2:, :] = np.random.sample((5, 7))
    for xi_tmp in test_pts:
        xi_tmp = np.random.sample(7)
        xi = {}
        xi["x1"] = int(1000 * xi_tmp[0])
        xi["x2"] = xi_tmp[1] - 1.0
        xi["x3"] = np.round(xi_tmp[2])
        xi["x4"] = f"{int(np.round(xi_tmp[3]))}guy"
        xi["x5"] = xi_tmp[4]
        xi["x6"] = xi_tmp[5] * 5.0
        xi["x7"] = xi_tmp[6]
        xxi = moop._embed(xi)
        # Check that embedding is legal
        assert (np.all(xxi[:6] >= -1.0e-8) and np.all(xxi[:6] <= 1 + 1.0e-8))
        assert (xxi[7] >= -1.0e-8 and xxi[7] <= 1 + 1.0e-8)
        assert (xxi.size == moop.n_latent)
        # Check extraction matches
        assert (moop._extract(xxi)["x1"] == xi["x1"])
        assert (np.abs(moop._extract(xxi)["x2"] - xi["x2"]) < 1.0e-8)
        assert (np.abs(moop._extract(xxi)["x3"] - xi["x3"]) < 1.0e-8)
        assert (moop._extract(xxi)["x4"] == xi["x4"])
        assert (moop._extract(xxi)["x5"] == xi["x5"])
        assert (np.abs(moop._extract(xxi)["x6"] - xi["x6"]) < 1.0e-8)
        assert (np.abs(moop._extract(xxi)["x7"] - xi["x7"]) < 1.0e-8)


def test_MOOP_pack_unpack_sim():
    """ Check that the MOOP class handles simulation packing correctly.

    Initialize a MOOP object with and without design variable names.
    Add 2 simulations and pack/unpack each output.

    """

    import numpy as np
    from parmoo import MOOP
    from parmoo.optimizers import LocalSurrogate_PS
    from parmoo.searches import LatinHypercube
    from parmoo.surrogates import GaussRBF

    # Create a continuous MOOP with 2 sims for packing/unpacking
    moop = MOOP(LocalSurrogate_PS)
    moop.addDesign({'name': "x1", 'lb': 0.0, 'ub': 1000.0},
                   {'name': "x2", 'lb': -1.0, 'ub': 0.0})
    g1 = {'m': 1,
          'hyperparams': {},
          'search': LatinHypercube,
          'sim_func': lambda x: [np.linalg.norm(x)],
          'surrogate': GaussRBF}
    g2 = {'m': 2,
          'search': LatinHypercube,
          'sim_func': lambda x: [np.linalg.norm(x-1.0), np.linalg.norm(x-0.5)],
          'surrogate': GaussRBF}
    moop.addSimulation(g1, g2)
    # Create a test vector
    sx = np.array([1.0, 2.0, 3.0])
    sxx = np.zeros(1, dtype=moop.sim_schema)
    sxx[0]['sim1'] = 1.0
    sxx[0]['sim2'][:] = np.array([2.0, 3.0])
    # Check packing
    assert (np.all(moop._pack_sim(sxx) == sx))
    # Check unpacking
    assert (moop._unpack_sim(sx)['sim1'] == sxx[0]['sim1'])
    assert (moop._unpack_sim(sx)['sim2'][0] == sxx[0]['sim2'][0])
    assert (moop._unpack_sim(sx)['sim2'][1] == sxx[0]['sim2'][1])


def test_MOOP_fit_update_surrogates():
    """ Check that the MOOP class handles evaluating surrogate models properly.

    Initialize a MOOP object and check that the _evaluate_surrogates() function
    works correctly.

    """

    from jax import config
    config.update("jax_enable_x64", True)
    import numpy as np
    from parmoo import MOOP
    from parmoo.acquisitions import UniformWeights
    from parmoo.optimizers import LocalSurrogate_PS
    from parmoo.searches import LatinHypercube
    from parmoo.surrogates import GaussRBF

    # Initialize a continuous MOOP with 2 sims, 3 objs
    moop1 = MOOP(LocalSurrogate_PS)
    for i in range(3):
        moop1.addDesign({'lb': 0.0, 'ub': 1.0})
    g1 = {'n': 3,
          'm': 1,
          'hyperparams': {},
          'search': LatinHypercube,
          'sim_func': lambda x: [np.linalg.norm([x[i] for i in x])],
          'surrogate': GaussRBF}
    g2 = {'n': 3,
          'm': 2,
          'hyperparams': {},
          'search': LatinHypercube,
          'sim_func': lambda x: [np.linalg.norm([x[i]-1 for i in x]),
                                 np.linalg.norm([x[i]-0.5 for i in x])],
          'surrogate': GaussRBF}
    moop1.addSimulation(g1, g2)
    moop1.addObjective({'obj_func': lambda x, s: x["x1"]},
                       {'obj_func': lambda x, s: s["sim1"]},
                       {'obj_func': lambda x, s: sum(s["sim2"])})
    moop1.addAcquisition({'acquisition': UniformWeights})
    moop1.compile()
    # Evaluate some data points and fit the surrogates
    for sn in ["sim1", "sim2"]:
        moop1.evaluateSimulation({'x1': 0, 'x2': 0, 'x3': 0}, sn)
        moop1.evaluateSimulation({'x1': 0.5, 'x2': 0.5, 'x3': 0.5}, sn)
        moop1.evaluateSimulation({'x1': 1, 'x2': 0, 'x3': 0}, sn)
        moop1.evaluateSimulation({'x1': 0, 'x2': 1, 'x3': 0}, sn)
        moop1.evaluateSimulation({'x1': 0, 'x2': 0, 'x3': 1}, sn)
        moop1.evaluateSimulation({'x1': 1, 'x2': 1, 'x3': 1}, sn)
    moop1._fit_surrogates()
    moop1._set_surrogate_tr(np.ones(3) * 0.5, np.ones(3) * 0.5)
    # Create an identical copy
    moop2 = MOOP(LocalSurrogate_PS)
    for i in range(3):
        moop2.addDesign({'lb': 0.0, 'ub': 1.0})
    moop2.addSimulation(g1, g2)
    moop2.addObjective({'obj_func': lambda x, s: x["x1"]},
                       {'obj_func': lambda x, s: s["sim1"]},
                       {'obj_func': lambda x, s: sum(s["sim2"])})
    moop2.addAcquisition({'acquisition': UniformWeights})
    moop2.compile()
    # Fit with half the training data used by moop1
    for sn in ["sim1", "sim2"]:
        moop2.evaluateSimulation({'x1': 0, 'x2': 0, 'x3': 0}, sn)
        moop2.evaluateSimulation({'x1': 0.5, 'x2': 0.5, 'x3': 0.5}, sn)
        moop2.evaluateSimulation({'x1': 1, 'x2': 0, 'x3': 0}, sn)
    moop2._fit_surrogates()
    moop2._set_surrogate_tr(np.ones(3) * 0.5, np.ones(3) * 0.5)
    # Update with the other half of the training data
    for sn in ["sim1", "sim2"]:
        moop2.evaluateSimulation({'x1': 0, 'x2': 1, 'x3': 0}, sn)
        moop2.evaluateSimulation({'x1': 0, 'x2': 0, 'x3': 1}, sn)
        moop2.evaluateSimulation({'x1': 1, 'x2': 1, 'x3': 1}, sn)
    moop2._update_surrogates()
    moop2._set_surrogate_tr(np.ones(3) * 0.5, np.ones(3) * 0.5)
    # Do 5 random tests and make sure the outputs are (near) identical
    for xi in np.random.sample((5, 3)):
        s1i = moop1._evaluate_surrogates(xi)
        s2i = moop2._evaluate_surrogates(xi)
        assert (np.linalg.norm(s1i - s2i) < 1.0e-8)


def test_MOOP_evaluate_surrogates():
    """ Check that the MOOP class handles evaluating surrogate models properly.

    Initialize a MOOP object and check that the _evaluate_surrogates() function
    works correctly.

    """

    from jax import config
    config.update("jax_enable_x64", True)
    import numpy as np
    from parmoo import MOOP
    from parmoo.acquisitions import UniformWeights
    from parmoo.optimizers import LocalSurrogate_PS
    from parmoo.searches import LatinHypercube
    from parmoo.surrogates import GaussRBF
    import pytest

    # Initialize a continuous MOOP with 2 sims, 3 objs
    moop1 = MOOP(LocalSurrogate_PS)
    for i in range(3):
        moop1.addDesign({'lb': 0.0, 'ub': 1.0})
    g1 = {'n': 3,
          'm': 1,
          'hyperparams': {},
          'search': LatinHypercube,
          'sim_func': lambda x: [np.linalg.norm([x[i] for i in x])],
          'surrogate': GaussRBF}
    g2 = {'n': 3,
          'm': 2,
          'hyperparams': {},
          'search': LatinHypercube,
          'sim_func': lambda x: [np.linalg.norm([x[i]-1 for i in x]),
                                 np.linalg.norm([x[i]-0.5 for i in x])],
          'surrogate': GaussRBF}
    moop1.addSimulation(g1, g2)
    moop1.addObjective({'obj_func': lambda x, s: x["x1"]},
                       {'obj_func': lambda x, s: s["sim1"]},
                       {'obj_func': lambda x, s: sum(s["sim2"])})
    moop1.addAcquisition({'acquisition': UniformWeights})
    moop1.compile()
    # Try some bad evaluations
    with pytest.raises(ValueError):
        moop1.evaluateSimulation(np.zeros(3), -1)
    # Evaluate some data points and fit the surrogates
    for sn in ["sim1", "sim2"]:
        moop1.evaluateSimulation({'x1': 0, 'x2': 0, 'x3': 0}, sn)
        moop1.evaluateSimulation({'x1': 0.5, 'x2': 0.5, 'x3': 0.5}, sn)
        moop1.evaluateSimulation({'x1': 1, 'x2': 0, 'x3': 0}, sn)
        moop1.evaluateSimulation({'x1': 0, 'x2': 1, 'x3': 0}, sn)
        moop1.evaluateSimulation({'x1': 0, 'x2': 0, 'x3': 1}, sn)
        moop1.evaluateSimulation({'x1': 1, 'x2': 1, 'x3': 1}, sn)
    moop1._fit_surrogates()
    moop1._set_surrogate_tr(np.ones(3) * 0.5, np.ones(3) * 0.5)
    # Now do some test evaluations and check the results
    test_cases = [
        (np.zeros(3), np.array([0, np.sqrt(3), np.sqrt(0.75)]), 0),
        (np.ones(3) / 2, np.array([np.sqrt(0.75), np.sqrt(0.75), 0]), 0),
        (np.eye(3)[0], np.array([1, np.sqrt(2), np.sqrt(0.75)]), 0),
        (np.eye(3)[1], np.array([1, np.sqrt(2),  np.sqrt(0.75)]), 0),
        (np.eye(3)[2], np.array([1, np.sqrt(2),  np.sqrt(0.75)]), 0),
        (np.ones(3), np.array([np.sqrt(3), 0.0, np.sqrt(0.75)]), 0)
    ]
    for xi, si, sdi in test_cases:
        assert (np.linalg.norm(moop1._evaluate_surrogates(xi) - si) < 1.e-7)
        assert (np.linalg.norm(moop1._surrogate_uncertainty(xi) - sdi) < 1.e-3)
    # Evaluate one point not in the training set and check that std_dev > 0
    xi = np.ones(3) * 0.75
    assert (np.linalg.norm(moop1._surrogate_uncertainty(xi)) > 1.0e-4)
    # Adjust the scale and try again
    moop2 = MOOP(LocalSurrogate_PS)
    moop2.addDesign({'lb': -1.0, 'ub': 1.0},
                    {'lb': 0.0, 'ub': 2.0},
                    {'lb': -0.5, 'ub': 1.5})
    moop2.addSimulation(g1, g2)
    moop2.addObjective({'obj_func': lambda x, s: x[0]},
                       {'obj_func': lambda x, s: s[0]},
                       {'obj_func': lambda x, s: s[1] + s[2]})
    moop2.addAcquisition({'acquisition': UniformWeights})
    moop2.compile()
    for sn in ["sim1", "sim2"]:
        moop2.evaluateSimulation({'x1': 0, 'x2': 0, 'x3': 0}, sn)
        moop2.evaluateSimulation({'x1': 0.5, 'x2': 0.5, 'x3': 0.5}, sn)
        moop2.evaluateSimulation({'x1': 1, 'x2': 0, 'x3': 0}, sn)
        moop2.evaluateSimulation({'x1': 0, 'x2': 1, 'x3': 0}, sn)
        moop2.evaluateSimulation({'x1': 0, 'x2': 0, 'x3': 1}, sn)
        moop2.evaluateSimulation({'x1': 1, 'x2': 1, 'x3': 1}, sn)
    moop2._fit_surrogates()
    moop2._set_surrogate_tr(np.zeros(3), np.inf)
    # Now compare evaluations against the original surrogate
    x = moop1._embed({'x1': 0, 'x2': 0, 'x3': 0})
    xx = moop2._embed({'x1': 0, 'x2': 0, 'x3': 0})
    assert (np.linalg.norm(moop1._evaluate_surrogates(x) -
                           moop2._evaluate_surrogates(xx)) < 1.0e-8)
    x = moop1._embed({'x1': 1, 'x2': 1, 'x3': 1})
    xx = moop2._embed({'x1': 1, 'x2': 1, 'x3': 1})
    assert (np.linalg.norm(moop1._evaluate_surrogates(x) -
                           moop2._evaluate_surrogates(xx)) < 1.0e-8)


def test_MOOP_evaluate_objectives():
    """ Check that the MOOP class handles evaluating objectives properly.

    Initialize a MOOP object and check that the _evaluate_objectives() function
    works correctly.

    """

    from jax import config
    config.update("jax_enable_x64", True)
    import numpy as np
    from parmoo import MOOP
    from parmoo.acquisitions import UniformWeights
    from parmoo.optimizers import LocalSurrogate_PS
    from parmoo.searches import LatinHypercube
    from parmoo.surrogates import GaussRBF
    import pytest

    # Initialize a continuous MOOP with 2 sims, 3 objs
    moop = MOOP(LocalSurrogate_PS)
    for i in range(3):
        moop.addDesign({'lb': 0.0, 'ub': 1.0})
    g1 = {'n': 3,
          'm': 1,
          'hyperparams': {},
          'search': LatinHypercube,
          'sim_func': lambda x: [np.linalg.norm([x[i] for i in x])],
          'surrogate': GaussRBF}
    g2 = {'n': 3,
          'm': 2,
          'hyperparams': {},
          'search': LatinHypercube,
          'sim_func': lambda x: [np.linalg.norm([x[i]-1 for i in x]),
                                 np.linalg.norm([x[i]-0.5 for i in x])],
          'surrogate': GaussRBF}
    moop.addSimulation(g1, g2)
    moop.addObjective({'obj_func': lambda x, s: x["x1"]},
                      {'obj_func': lambda x, s: s["sim1"][0]},
                      {'obj_func': lambda x, s: s["sim2"][0] + s["sim2"][1]})
    moop.addAcquisition({'acquisition': UniformWeights})
    moop.compile()
    # Try some bad evaluations
    with pytest.raises(ValueError):
        moop.evaluateSimulation(np.zeros(3), -1)
    # Evaluate some data points and fit the surrogates
    for sn in ["sim1", "sim2"]:
        moop.evaluateSimulation({'x1': 0, 'x2': 0, 'x3': 0}, sn)
        moop.evaluateSimulation({'x1': 0.5, 'x2': 0.5, 'x3': 0.5}, sn)
        moop.evaluateSimulation({'x1': 1, 'x2': 0, 'x3': 0}, sn)
        moop.evaluateSimulation({'x1': 0, 'x2': 1, 'x3': 0}, sn)
        moop.evaluateSimulation({'x1': 0, 'x2': 0, 'x3': 1}, sn)
        moop.evaluateSimulation({'x1': 1, 'x2': 1, 'x3': 1}, sn)
    moop._fit_surrogates()
    moop._set_surrogate_tr(np.ones(3) * 0.5, np.ones(3) * 0.5)
    # Now do some test evaluations and check the results
    test_cases = [
        (np.zeros(3), np.array([0.0, 0.0, np.sqrt(3) + np.sqrt(0.75)])),
        (np.ones(3) * 0.5, np.array([0.5, np.sqrt(0.75), np.sqrt(0.75)])),
        (np.eye(3)[0], np.array([1.0, 1.0, np.sqrt(2) + np.sqrt(0.75)])),
        (np.eye(3)[1], np.array([0.0, 1.0, np.sqrt(2) + np.sqrt(0.75)])),
        (np.eye(3)[2], np.array([0.0, 1.0, np.sqrt(2) + np.sqrt(0.75)])),
        (np.ones(3), np.array([1.0, np.sqrt(3), np.sqrt(0.75)]))
    ]
    for xi, fi in test_cases:
        sxi = moop._evaluate_surrogates(xi)
        fxi = moop._evaluate_objectives(xi, sxi)
        assert (np.linalg.norm(fi - fxi) < 1.0e-8)


def test_MOOP_evaluate_constraints():
    """ Check that the MOOP class handles evaluating constraints properly.

    Initialize a MOOP object and check that the _evaluate_constraints()
    function works correctly.

    """

    from jax import config
    config.update("jax_enable_x64", True)
    import numpy as np
    from parmoo import MOOP
    from parmoo.acquisitions import UniformWeights
    from parmoo.optimizers import LocalSurrogate_PS
    from parmoo.searches import LatinHypercube
    from parmoo.surrogates import GaussRBF

    # Initialize a continuous MOOP with 2 sims, 3 cons
    moop = MOOP(LocalSurrogate_PS)
    for i in range(3):
        moop.addDesign({'lb': 0.0, 'ub': 1.0})
    g1 = {'n': 3,
          'm': 1,
          'hyperparams': {},
          'search': LatinHypercube,
          'sim_func': lambda x: [np.linalg.norm([x[i] for i in x])],
          'surrogate': GaussRBF}
    g2 = {'n': 3,
          'm': 2,
          'hyperparams': {},
          'search': LatinHypercube,
          'sim_func': lambda x: [np.linalg.norm([x[i]-1 for i in x]),
                                 np.linalg.norm([x[i]-0.5 for i in x])],
          'surrogate': GaussRBF}
    moop.addSimulation(g1, g2)
    moop.addObjective({'obj_func': lambda x, s: x["x1"]})
    moop.addAcquisition({'acquisition': UniformWeights})
    moop.compile()
    # Evaluate an empty constraint and check that a zero array is returned
    assert (np.all(moop._evaluate_constraints(np.zeros(3), np.zeros(3))
            == np.zeros(1)))
    # Now add 3 constraints
    moop.addConstraint({'constraint': lambda x, s: x["x1"]})
    moop.addConstraint({'constraint': lambda x, s: s["sim1"][0]})
    moop.addConstraint({'constraint':
                        lambda x, s: s["sim2"][0] + s["sim2"][1]})
    moop.compile()
    # Evaluate some data points and fit the surrogates
    for sn in ["sim1", "sim2"]:
        moop.evaluateSimulation({'x1': 0, 'x2': 0, 'x3': 0}, sn)
        moop.evaluateSimulation({'x1': 0.5, 'x2': 0.5, 'x3': 0.5}, sn)
        moop.evaluateSimulation({'x1': 1, 'x2': 0, 'x3': 0}, sn)
        moop.evaluateSimulation({'x1': 0, 'x2': 1, 'x3': 0}, sn)
        moop.evaluateSimulation({'x1': 0, 'x2': 0, 'x3': 1}, sn)
        moop.evaluateSimulation({'x1': 1, 'x2': 1, 'x3': 1}, sn)
    moop._fit_surrogates()
    moop._set_surrogate_tr(np.zeros(3), np.inf)
    # Now do some test evaluations and check the results
    test_cases = [
        (np.zeros(3), np.array([0.0, 0.0, np.sqrt(3) + np.sqrt(0.75)])),
        (np.ones(3) * 0.5, np.array([0.5, np.sqrt(0.75), np.sqrt(0.75)])),
        (np.eye(3)[0], np.array([1.0, 1.0, np.sqrt(2) + np.sqrt(0.75)])),
        (np.eye(3)[1], np.array([0.0, 1.0, np.sqrt(2) + np.sqrt(0.75)])),
        (np.eye(3)[2], np.array([0.0, 1.0, np.sqrt(2) + np.sqrt(0.75)])),
        (np.ones(3), np.array([1.0, np.sqrt(3), np.sqrt(0.75)]))
    ]
    for xi, ci in test_cases:
        sxi = moop._evaluate_surrogates(xi)
        cxi = moop._evaluate_constraints(xi, sxi)
        assert (np.linalg.norm(ci - cxi) < 1.0e-8)


def test_MOOP_evaluate_penalty():
    """ Check that the MOOP class handles evaluating penalty function properly.

    Initialize a MOOP object and check that the _evaluate_penalty() function
    works correctly.

    """

    from jax import config
    config.update("jax_enable_x64", True)
    import numpy as np
    from parmoo import MOOP
    from parmoo.acquisitions import UniformWeights
    from parmoo.optimizers import LocalSurrogate_PS
    from parmoo.searches import LatinHypercube
    from parmoo.surrogates import GaussRBF
    import pytest

    # Initialize a continuous MOOP with 2 sims, 3 objs, 1 cons
    moop = MOOP(LocalSurrogate_PS)
    for i in range(3):
        moop.addDesign({'lb': 0.0, 'ub': 1.0})
    g1 = {'n': 3,
          'm': 1,
          'hyperparams': {},
          'search': LatinHypercube,
          'sim_func': lambda x: [np.linalg.norm([x[i] for i in x])],
          'surrogate': GaussRBF}
    g2 = {'n': 3,
          'm': 2,
          'hyperparams': {},
          'search': LatinHypercube,
          'sim_func': lambda x: [np.linalg.norm([x[i]-1 for i in x]),
                                 np.linalg.norm([x[i]-0.5 for i in x])],
          'surrogate': GaussRBF}
    moop.addSimulation(g1, g2)
    moop.addObjective({'obj_func': lambda x, s: x["x1"]},
                      {'obj_func': lambda x, s: s["sim1"][0]},
                      {'obj_func': lambda x, s: s["sim2"][0] + s["sim2"][1]})
    moop.addConstraint({'constraint': lambda x, s: x["x1"] - 0.5})
    moop.addAcquisition({'acquisition': UniformWeights})
    moop.compile()
    # Try some bad evaluations
    with pytest.raises(ValueError):
        moop.evaluateSimulation(np.zeros(3), -1)
    # Evaluate some data points and fit the surrogates
    for sn in ["sim1", "sim2"]:
        moop.evaluateSimulation({'x1': 0, 'x2': 0, 'x3': 0}, sn)
        moop.evaluateSimulation({'x1': 0.5, 'x2': 0.5, 'x3': 0.5}, sn)
        moop.evaluateSimulation({'x1': 1, 'x2': 0, 'x3': 0}, sn)
        moop.evaluateSimulation({'x1': 0, 'x2': 1, 'x3': 0}, sn)
        moop.evaluateSimulation({'x1': 0, 'x2': 0, 'x3': 1}, sn)
        moop.evaluateSimulation({'x1': 1, 'x2': 1, 'x3': 1}, sn)
    moop._fit_surrogates()
    moop._set_surrogate_tr(np.ones(3) * 0.5, np.ones(3) * 0.5)
    # Now do some test evaluations and check the results
    test_cases = [
        (np.zeros(3), np.array([0.0, 0.0, np.sqrt(3) + np.sqrt(0.75)])),
        (np.ones(3) * 0.5, np.array([0.5, np.sqrt(0.75), np.sqrt(0.75)])),
        (np.eye(3)[0], np.array([1.0, 1.0, np.sqrt(2) + np.sqrt(0.75)]) + 0.5),
        (np.eye(3)[1], np.array([0.0, 1.0, np.sqrt(2) + np.sqrt(0.75)])),
        (np.eye(3)[2], np.array([0.0, 1.0, np.sqrt(2) + np.sqrt(0.75)])),
        (np.ones(3), np.array([1.0, np.sqrt(3), np.sqrt(0.75)]) + 0.5)
    ]
    for xi, pi in test_cases:
        sxi = moop._evaluate_surrogates(xi)
        pxi = moop._evaluate_penalty(xi, sxi)
        assert (np.linalg.norm(pi - pxi) < 1.0e-8)


if __name__ == "__main__":
    test_MOOP_embed_extract()
    test_MOOP_pack_unpack_sim()
    test_MOOP_fit_update_surrogates()
    test_MOOP_evaluate_surrogates()
    test_MOOP_evaluate_objectives()
    test_MOOP_evaluate_constraints()
    test_MOOP_evaluate_penalty()




""" Use ParMOO to solve a convex, user-defined problem, with mixed variables
(continuous, integer, categorical, and custom).

Uses named variables and public function definitions to define the problem.

"""

import numpy as np
import os
from parmoo import MOOP
from parmoo.optimizers import GlobalSurrogate_PS
from parmoo.surrogates import GaussRBF
from parmoo.acquisitions import RandomConstraint
from parmoo.searches import LatinHypercube
from parmoo.embeddings import ContinuousEmbedder

# For this test, use all user-defined functions

def sim(x):
    " User sim for sample problem. "
    # Categorical variables
    if x["cat var 1"] == "good":
        result = 0.0
    else:
        result = 1.0
    if x["cat var 2"] != "low":
        result += 1.0
    return np.array([result + x["cont var"] ** 2 + x["int var"] ** 2 +
                     float(x["custom var"]) ** 2,
                     result + (x["cont var"] - 1.0) ** 2 + x["int var"] ** 2 +
                     float(x["custom var"]) ** 2])

def obj1(x, sx):
    " User obj1 for sample problem. "
    return sx["my sim"][0]

def obj2(x, sx):
    " User obj2 for sample problem. "
    return sx["my sim"][1]

# Create a MOOP
moop = MOOP(GlobalSurrogate_PS)

# Add design variables
moop.addDesign({'name': "cont var", 'ub': 1.0, 'lb': 0.0,
                'des_type': "continuous", 'des_tol': 1.0e-8})
moop.addDesign({'name': "int var", 'ub': 5, 'lb': -5,
                'des_type': "integer"})
moop.addDesign({'name': "cat var 1", 'des_type': "categorical",
                'levels': ["good", "bad"]})
moop.addDesign({'name': "cat var 2", 'des_type': "categorical",
                'levels': ["low", "med", "high"]})
moop.addDesign({'name': "custom var", 'des_type': "custom",
                'ub': 1.0, 'lb': 0.0,
                'embedding_size': 1,
                'embedder': ContinuousEmbedder})

# Add the simulation
moop.addSimulation({'name': "my sim",
                    'm': 2,
                    'sim_func': sim,
                    'search': LatinHypercube,
                    'surrogate': GaussRBF,
                    'hyperparams': {'search_budget': 20}})

# Add user objective functions
moop.addObjective({'obj_func': obj1}, {'obj_func': obj2})

# Add 1 single acquisition function
for i in range(1):
    moop.addAcquisition({'acquisition': RandomConstraint, 'hyperparams': {}})

# Solve the problem with 10 iterations with checkpointing on
moop.solve(10)

# Check that 30 simulations were evaluated and solutions are feasible
assert(moop.getSimulationData()["my sim"].size == 30)
assert(moop.getObjectiveData().size == 30)
assert(moop.getPF().size > 0)




def test_GaussRBF():
    """ Test the GaussRBF class in surrogates.py.

    Generate random x and y values for a 3D input space and 2D output space,
    then use the GaussRBF class to fit Gaussian RBFs to the data.

    Check that the Gaussian RBFs interpolate the original data.

    """

    from jax import config
    config.update("jax_enable_x64", True)
    from jax import jacrev
    import numpy as np
    import os
    from parmoo.surrogates import GaussRBF
    import pytest

    # Try some bad initializations to test error handling
    with pytest.raises(ValueError):
        GaussRBF(2, np.zeros(3), np.ones(3), {'nugget': []})
    with pytest.raises(ValueError):
        GaussRBF(2, np.zeros(3), np.ones(3), {'nugget': -1.0})
    with pytest.raises(ValueError):
        GaussRBF(2, np.zeros(3), np.ones(3),
                 {'nugget': 0.1, 'des_tols': np.zeros(3)})
    with pytest.raises(ValueError):
        GaussRBF(2, np.zeros(3), np.ones(3),
                 {'nugget': 0.1, 'des_tols': np.zeros(2)})
    with pytest.raises(ValueError):
        GaussRBF(2, np.zeros(3), np.ones(3),
                 {'nugget': 0.1, 'des_tols': 0.1})
    # Create 2 identical RBFs
    rbf1 = GaussRBF(2, np.zeros(3), np.ones(3), {})
    rbf2 = GaussRBF(2, np.zeros(3), np.ones(3), {})
    # Generate some random data with 3 design variables and 2 outputs
    x_vals1 = np.random.random_sample((10, 3))
    y_vals1 = np.random.random_sample((10, 2))
    x_vals2 = np.random.random_sample((10, 3))
    y_vals2 = np.random.random_sample((10, 2))
    x_vals_full = np.concatenate((x_vals1, x_vals2), axis=0)
    y_vals_full = np.concatenate((y_vals1, y_vals2), axis=0)
    # Try to fit/update with illegal data to test error handling
    with pytest.raises(TypeError):
        rbf1.fit(0, y_vals1)
    with pytest.raises(TypeError):
        rbf1.fit(x_vals1, 0)
    with pytest.raises(ValueError):
        rbf1.fit(np.zeros((0, 3)), np.zeros((0, 2)))
    with pytest.raises(ValueError):
        rbf1.fit(np.zeros((10, 3)), np.zeros((10, 3)))
    with pytest.raises(ValueError):
        rbf1.fit(np.zeros((10, 3)), np.zeros((9, 2)))
    with pytest.raises(TypeError):
        rbf1.update(0, y_vals1)
    with pytest.raises(TypeError):
        rbf1.update(x_vals1, 0)
    with pytest.raises(ValueError):
        rbf1.update(np.zeros((10, 3)), np.zeros((10, 3)))
    with pytest.raises(ValueError):
        rbf1.update(np.zeros((10, 3)), np.zeros((9, 2)))
    # Fit both RBFs, with and without using the update method
    rbf1.fit(x_vals1, y_vals1)
    rbf1.update(x_vals2, y_vals2)
    rbf1.update(np.zeros((0, 3)), np.zeros((0, 2)))    # Update with no data
    rbf1.setTrustRegion(np.zeros(3), np.ones(3) * np.inf)
    rbf2.fit(x_vals_full, y_vals_full)
    rbf2.setTrustRegion(np.zeros(3), np.ones(3) * np.inf)
    # Try a bad improvement step to test error handling
    with pytest.raises(TypeError):
        rbf1.improve(5, False)
    with pytest.raises(ValueError):
        rbf1.improve(np.zeros(2), False)
    with pytest.raises(ValueError):
        rbf1.improve(-np.ones(3), False)
    # Check that the RBFs match on a random evaluation point
    x = np.random.random_sample((3))
    assert (all(rbf1.evaluate(x) == rbf2.evaluate(x)))
    assert (all(rbf1.stdDev(x) == rbf2.stdDev(x)))
    # Check that the RBFs interpolate, up to 8 decimal digits of precision
    for i in range(x_vals_full.shape[0]):
        assert (np.linalg.norm(rbf1.evaluate(x_vals_full[i]) - y_vals_full[i])
                < 1.0e-8)
        assert (np.linalg.norm(rbf2.evaluate(x_vals_full[i]) - y_vals_full[i])
                < 1.0e-8)
        assert (np.max(rbf1.stdDev(x_vals_full[i])) < 1.0e-4)
        assert (np.max(rbf2.stdDev(x_vals_full[i])) < 1.0e-4)
    # Check that the RBFs compute the same grad, up to 8 digits of precision
    for i in range(x_vals_full.shape[0]):
        assert (np.linalg.norm(jacrev(rbf1.evaluate)(x_vals_full[i]) -
                               jacrev(rbf2.evaluate)(x_vals_full[i]))
                < 1.0e-8)
    for i in range(x_vals_full.shape[0]):
        assert (np.linalg.norm(jacrev(rbf1.stdDev)(x_vals_full[i]) -
                               jacrev(rbf2.stdDev)(x_vals_full[i])) < 1.0e-4)
    # Check that the RBF gradient evaluates correctly on a known dataset
    x_vals3 = np.eye(3)
    x_vals3 = np.append(x_vals3, [[0.5, 0.5, 0.5]], axis=0)
    y_vals3 = np.asarray([[np.dot(xi, xi)] for xi in x_vals3])
    rbf3 = GaussRBF(1, np.zeros(3), np.ones(3), {})
    rbf3.fit(x_vals3, y_vals3)
    rbf3.setTrustRegion(np.zeros(3), np.ones(3) * np.inf)
    y_grad_vals3 = -0.03661401 * np.ones((1, 3))
    assert (np.linalg.norm(jacrev(rbf3.evaluate)(x_vals3[-1])
                           - y_grad_vals3[-1]) < 1.0e-4)
    # Check standard deviation calculations
    xi = np.random.random_sample(3)
    assert (np.all(rbf3.stdDev(xi) >= 0))
    assert (np.any(jacrev(rbf3.stdDev)(xi) != 0))
    # Check that the RBF generates feasible local improvement points
    for i in range(4):
        x_improv = rbf3.improve(np.zeros(3), False)
        assert (np.all(x_improv[0] <= np.ones(3)) and
                np.all(x_improv[0] >= np.zeros(3)))
    # Check that the RBF generates feasible global improvement points
    x_improv = rbf3.improve(x_vals3[-1], True)
    assert (np.all(x_improv[0] <= np.ones(3)) and
            np.all(x_improv[0] >= np.zeros(3)))
    # Check that RBF generates good improvements when points are bunched
    x_new = np.ones((3, 3))
    x_new = x_new * 0.5
    f_new = np.zeros((3, 1))
    for i in range(3):
        x_new[i, i] = 0.50000001
        f_new[i, 0] = np.dot(x_new[i, :], x_new[i, :])
    rbf3.update(x_new, f_new)
    x_improv = rbf3.improve(np.asarray([0.5, 0.5, 0.5]), False)
    assert (np.all(x_improv[0] <= np.ones(3)) and
            np.all(x_improv[0] >= np.zeros(3)))
    assert (np.all(np.linalg.norm(x_improv[0] - np.asarray([0.5, 0.5, 0.5]))
            > 1.0e-8))
    # Now fit redundant data points using a nugget
    x_vals1 = np.append(x_vals1, np.asarray([x_vals1[0, :]]), axis=0)
    y_vals1 = np.append(y_vals1, np.asarray([y_vals1[0, :]]), axis=0)
    rbf4 = GaussRBF(2, np.zeros(3), np.ones(3), {'nugget': 0.0001})
    rbf4.fit(x_vals1, y_vals1)
    rbf4.update(x_vals2, y_vals2)
    rbf4.setTrustRegion(np.zeros(3), np.ones(3) * np.inf)
    # Now create a really tiny design space with a large tolerance
    rbf5 = GaussRBF(1, np.zeros(1), np.ones(1),
                    {'des_tols': 0.3 * np.ones(1)})
    xdat5 = np.asarray([[0.4], [0.6]])
    ydat5 = np.asarray([[0.4], [0.6]])
    rbf5.fit(xdat5, ydat5)
    rbf5.setTrustRegion(np.zeros(1), np.ones(1) * np.inf)
    # Test that improve() is able to find points outside the design tolerance
    for i in range(5):
        x_improv = rbf5.improve(xdat5[0], False)
        assert (x_improv[0][0] < 0.1 or x_improv[0][0] > 0.9)
    for i in range(5):
        x_improv = rbf5.improve(xdat5[0], True)
        assert (x_improv[0][0] < 0.1 or x_improv[0][0] > 0.9)
    # Now fit redundant data points without a nugget
    rbf6 = GaussRBF(2, np.zeros(3), np.ones(3), {})
    rbf6.fit(x_vals1, y_vals1)
    rbf6.update(x_vals2, y_vals2)
    rbf6.setTrustRegion(np.zeros(3), np.ones(3) * np.inf)
    # Now fit datapoints in a plane
    x_vals3 = np.zeros((4, 3))
    x_vals3[1, 0] = 0.1
    x_vals3[2, 0] = 0.2
    x_vals3[3, 0] = 0.3
    y_vals3 = np.ones((4, 2))
    rbf7 = GaussRBF(2, np.zeros(3), np.ones(3), {})
    rbf7.fit(x_vals3, y_vals3)
    rbf7.update(x_vals3, y_vals3)
    rbf7.setTrustRegion(np.zeros(3), np.ones(3) * np.inf)
    # Test save and load
    rbf6.save("parmoo.surrogate")
    rbf7.load("parmoo.surrogate")
    os.remove("parmoo.surrogate")
    rbf7.setTrustRegion(np.zeros(3), np.ones(3) * np.inf)
    xx = np.random.random_sample(3)
    assert (np.all(rbf6.evaluate(xx) == rbf7.evaluate(xx)))
    # Generate a simple 1D RBF and check that its stdDev is accurate
    x_vals4 = np.array([[0.0], [1.0]])
    y_vals4 = np.array([[0.0], [1.0]])
    rbf8 = GaussRBF(1, np.zeros(1), np.ones(1), {'tail_order': 0})
    rbf8.fit(x_vals4, y_vals4)
    rbf8.setTrustRegion(np.zeros(1), np.ones(1) * np.inf)
    assert (np.linalg.norm(rbf8.evaluate(np.array([0.5])) - 0.5) < 1.0e-8)
    assert (np.linalg.norm(rbf8.stdDev(np.array([0.5]))) > 1.0e-2)
    assert (np.linalg.norm(jacrev(rbf8.evaluate)(np.array([0.5]))) > 1)
    assert (np.linalg.norm(jacrev(rbf8.stdDev)(np.array([0.5]))) < 1.0e-4)
    xx = np.linspace(0, 1).reshape((50, 1))
    maxind = 0
    for i, xi in enumerate(xx):
        if np.all(rbf8.stdDev(xi) > rbf8.stdDev(xx[maxind])):
            maxind = i
        if i < 25:
            assert (np.all(jacrev(rbf8.stdDev)(xi) >= 0))
        else:
            assert (np.all(jacrev(rbf8.stdDev)(xi) <= 0))
    assert (maxind in [24, 25])
    return


def test_LocalGaussRBF():
    """ Test the LocalGaussRBF class in surrogates.py.

    Generate random x and y values for a 3D input space and 2D output space,
    then use the LocalGaussRBF class to fit Gaussian RBFs to the data.

    Check that the local Gaussian RBFs interpolate the original data.

    """

    from jax import config
    config.update("jax_enable_x64", True)
    from jax import jacrev
    import numpy as np
    import os
    from parmoo.surrogates import GaussRBF
    import pytest

    # Try some bad initializations to test error handling
    with pytest.raises(ValueError):
        GaussRBF(2, np.zeros(3), np.ones(3), {'nugget': []})
    with pytest.raises(ValueError):
        GaussRBF(2, np.zeros(3), np.ones(3), {'nugget': -1.0})
    with pytest.raises(ValueError):
        GaussRBF(2, np.zeros(3), np.ones(3),
                 {'nugget': 0.1, 'des_tols': np.zeros(3)})
    with pytest.raises(ValueError):
        GaussRBF(2, np.zeros(3), np.ones(3),
                 {'nugget': 0.1, 'des_tols': np.zeros(2)})
    with pytest.raises(ValueError):
        GaussRBF(2, np.zeros(3), np.ones(3), {'nugget': 0.1, 'des_tols': 0.1})
    # Create 2 identical RBFs
    rbf1 = GaussRBF(2, np.zeros(3), np.ones(3), {'tail_order': 1})
    rbf2 = GaussRBF(2, np.zeros(3), np.ones(3), {'tail_order': 1})
    # Generate some random data with 3 design variables and 2 outputs
    x_vals1 = np.random.random_sample((10, 3))
    y_vals1 = np.random.random_sample((10, 2))
    x_vals2 = np.random.random_sample((10, 3))
    y_vals2 = np.random.random_sample((10, 2))
    x_vals_full = np.concatenate((x_vals1, x_vals2), axis=0)
    y_vals_full = np.concatenate((y_vals1, y_vals2), axis=0)
    # Try to fit/update with illegal data to test error handling
    with pytest.raises(TypeError):
        rbf1.fit(0, y_vals1)
    with pytest.raises(TypeError):
        rbf1.fit(x_vals1, 0)
    with pytest.raises(ValueError):
        rbf1.fit(np.zeros((0, 3)), np.zeros((0, 2)))
    with pytest.raises(ValueError):
        rbf1.fit(np.zeros((10, 3)), np.zeros((10, 3)))
    with pytest.raises(ValueError):
        rbf1.fit(np.zeros((10, 3)), np.zeros((9, 2)))
    with pytest.raises(TypeError):
        rbf1.update(0, y_vals1)
    with pytest.raises(TypeError):
        rbf1.update(x_vals1, 0)
    with pytest.raises(ValueError):
        rbf1.update(np.zeros((10, 3)), np.zeros((10, 3)))
    with pytest.raises(ValueError):
        rbf1.update(np.zeros((10, 3)), np.zeros((9, 2)))
    # Fit both RBFs, with and without using the update method
    rbf1.fit(x_vals1, y_vals1)
    rbf1.update(x_vals2, y_vals2)
    rbf1.update(np.zeros((0, 3)), np.zeros((0, 2)))    # Update with no data
    rbf1.setTrustRegion(0.5 * np.ones(3), np.ones(3) * 0.5)
    rbf2.fit(x_vals_full, y_vals_full)
    rbf2.setTrustRegion(0.5 * np.ones(3), np.ones(3) * 0.5)
    # Try to set the center with illegal values to test error handling
    with pytest.raises(TypeError):
        rbf1.setTrustRegion(5, np.ones(3))
    with pytest.raises(ValueError):
        rbf1.setTrustRegion(np.zeros(5), np.ones(3))
    with pytest.raises(ValueError):
        rbf1.setTrustRegion(-np.ones(3), np.ones(3))
    with pytest.raises(TypeError):
        rbf1.setTrustRegion(np.ones(3) * 0.5, 5)
    with pytest.raises(ValueError):
        rbf1.setTrustRegion(np.ones(3) * 0.5, np.ones(5))
    with pytest.raises(ValueError):
        rbf1.setTrustRegion(np.ones(3) * 0.5, -np.ones(3))
    # Try a bad improvement step to test error handling
    with pytest.raises(TypeError):
        rbf1.improve(5, False)
    with pytest.raises(ValueError):
        rbf1.improve(np.zeros(2), False)
    with pytest.raises(ValueError):
        rbf1.improve(-np.ones(3), False)
    # Check that the RBFs match on a random evaluation point
    x = np.random.random_sample((3))
    assert (all(rbf1.evaluate(x) == rbf2.evaluate(x)))
    assert (all(rbf1.stdDev(x) == rbf2.stdDev(x)))
    # Check that the RBFs interpolate, up to 8 decimal digits of precision
    for i in range(x_vals_full.shape[0]):
        rbf1.setTrustRegion(x_vals_full[i], np.ones(3) * 0.1)
        rbf2.setTrustRegion(x_vals_full[i], np.ones(3) * 0.1)
        assert (np.linalg.norm(rbf1.evaluate(x_vals_full[i])-y_vals_full[i])
                < 1.0e-8)
        assert (np.linalg.norm(rbf2.evaluate(x_vals_full[i])-y_vals_full[i])
                < 1.0e-8)
        assert (np.linalg.norm(rbf1.stdDev(x_vals_full[i]) < 1.0e-4))
        assert (np.linalg.norm(rbf2.stdDev(x_vals_full[i]) < 1.0e-4))
    # Check that the RBFs compute the same grad, up to 8 digits of precision
    rbf1.setTrustRegion(0.5 * np.ones(3), np.ones(3) * 0.1)
    rbf2.setTrustRegion(0.5 * np.ones(3), np.ones(3) * 0.1)
    for i in range(x_vals_full.shape[0]):
        assert (np.linalg.norm(jacrev(rbf1.evaluate)(x_vals_full[i]) -
                               jacrev(rbf2.evaluate)(x_vals_full[i])) < 1.0e-8)
        assert (np.linalg.norm(jacrev(rbf1.stdDev)(x_vals_full[i]) -
                               jacrev(rbf2.stdDev)(x_vals_full[i])) < 1.0e-4)
    # Check that the RBF gradient evaluates correctly on a known dataset
    x_vals3 = np.eye(3)
    x_vals3 = np.append(x_vals3, [[0.5, 0.5, 0.5]], axis=0)
    y_vals3 = np.asarray([[np.dot(xi, xi)] for xi in x_vals3])
    rbf3 = GaussRBF(1, np.zeros(3), np.ones(3), {'tail_order': 0})
    rbf3.fit(x_vals3, y_vals3)
    rbf3.setTrustRegion(x_vals3[-1], np.ones(3) * 0.25)
    y_grad_vals3 = -0.08798618 * np.ones((1, 3))
    assert (np.linalg.norm(jacrev(rbf3.evaluate)(x_vals3[-1])
                           - y_grad_vals3[-1]) < 1.0e-4)
    assert (np.linalg.norm(jacrev(rbf3.stdDev)(x_vals3[-1]) >= 0))
    # Check standard deviation calculations
    xi = np.random.random_sample(3)
    assert (np.all(rbf3.stdDev(xi) >= 0))
    assert (np.any(jacrev(rbf3.stdDev)(xi) != 0))
    # Check that the RBF generates feasible local improvement points
    for i in range(4):
        x_improv = rbf3.improve(np.zeros(3), False)
        assert (np.all(x_improv[0] <= np.ones(3)) and
                np.all(x_improv[0] >= np.zeros(3)))
    # Check that the RBF generates feasible global improvement points
    x_improv = rbf3.improve(x_vals3[-1], True)
    assert (np.all(x_improv[0] <= np.ones(3)) and
            np.all(x_improv[0] >= np.zeros(3)))
    # Check that RBF generates good improvements when points are bunched
    x_new = np.ones((3, 3))
    x_new = x_new * 0.5
    f_new = np.zeros((3, 1))
    for i in range(3):
        x_new[i, i] = 0.50000001
        f_new[i, 0] = np.dot(x_new[i, :], x_new[i, :])
    rbf3.update(x_new, f_new)
    rbf3.setTrustRegion(0.5 * np.ones(3), 1.0e-4)
    x_improv = rbf3.improve(np.asarray([0.5, 0.5, 0.5]), False)
    assert (np.all(x_improv[0] <= np.ones(3)) and
            np.all(x_improv[0] >= np.zeros(3)))
    assert (np.all(np.linalg.norm(x_improv[0] - np.asarray([0.5, 0.5, 0.5]))
            > 1.0e-8))
    # Now fit redundant data points using a nugget
    x_vals1 = np.append(x_vals1, np.asarray([x_vals1[0, :]]), axis=0)
    y_vals1 = np.append(y_vals1, np.asarray([y_vals1[0, :]]), axis=0)
    rbf4 = GaussRBF(2, np.zeros(3), np.ones(3), {'nugget': 0.0001})
    rbf4.fit(x_vals1, y_vals1)
    rbf4.setTrustRegion(x_vals1[0], np.ones(3) * 0.1)
    rbf4.evaluate(x_vals1[0])
    rbf4.update(x_vals2, y_vals2)
    rbf4.setTrustRegion(x_vals1[0], np.ones(3) * 0.1)
    jacrev(rbf4.evaluate)(x_vals1[0])
    # Now create a really tiny design space with a large tolerance
    rbf5 = GaussRBF(1, np.zeros(1), np.ones(1), {'des_tols': 0.3 * np.ones(1)})
    xdat5 = np.asarray([[0.4], [0.6]])
    ydat5 = np.asarray([[0.4], [0.6]])
    rbf5.fit(xdat5, ydat5)
    rbf5.setTrustRegion(np.ones(1) * 0.5, np.ones(1) * 1.0e-1)
    # Test that improve() is able to find points outside the design tolerance
    for i in range(5):
        x_improv = rbf5.improve(xdat5[0], False)
        assert (x_improv[0][0] < 0.1 or x_improv[0][0] > 0.9)
    for i in range(5):
        x_improv = rbf5.improve(xdat5[0], True)
        assert (x_improv[0][0] < 0.1 or x_improv[0][0] > 0.9)
    # Now fit redundant data points to test adaptive nugget
    rbf6 = GaussRBF(2, np.zeros(3), np.ones(3), {})
    rbf6.fit(x_vals1, y_vals1)
    rbf6.update(x_vals2, y_vals2)
    rbf6.setTrustRegion(x_vals1[0], np.ones(3) * 0.1)
    # Now fit datapoints in a plane
    x_vals3 = np.zeros((4, 3))
    x_vals3[1, 0] = 0.1
    x_vals3[2, 0] = 0.2
    x_vals3[3, 0] = 0.3
    y_vals3 = np.ones((4, 2))
    rbf7 = GaussRBF(2, np.zeros(3), np.ones(3), {})
    rbf7.fit(x_vals3, y_vals3)
    rbf7.update(x_vals3, y_vals3)
    # Test save and load
    rbf6.save("parmoo.surrogate")
    rbf7.load("parmoo.surrogate")
    os.remove("parmoo.surrogate")
    rbf7.setTrustRegion(np.zeros(3), np.ones(3) * np.inf)
    xx = np.random.random_sample(3)
    assert (np.all(rbf6.evaluate(xx) == rbf7.evaluate(xx)))
    # Generate a simple 1D RBF and check that its stdDev is accurate
    x_vals4 = np.array([[0.0], [1.0]])
    y_vals4 = np.array([[0.0], [1.0]])
    rbf8 = GaussRBF(1, np.zeros(1), np.ones(1), {'tail_order': 0})
    rbf8.fit(x_vals4, y_vals4)
    rbf8.setTrustRegion(np.array([0.5]), np.ones(1) * 0.25)
    assert (np.linalg.norm(rbf8.evaluate(np.array([0.5])) - 0.5) < 1.0e-8)
    assert (np.linalg.norm(rbf8.stdDev(np.array([0.5]))) > 5.0e-3)
    assert (np.linalg.norm(jacrev(rbf8.evaluate)(np.array([0.5]))) > 1.0)
    assert (np.linalg.norm(jacrev(rbf8.stdDev)(np.array([0.5]))) < 1.0e-4)
    xx = np.linspace(0, 1).reshape((50, 1))
    maxind = 0
    for i, xi in enumerate(xx):
        if np.all(rbf8.stdDev(xi) > rbf8.stdDev(xx[maxind])):
            maxind = i
        if i < 25:
            assert (np.all(jacrev(rbf8.stdDev)(xi) >= 0))
        else:
            assert (np.all(jacrev(rbf8.stdDev)(xi) <= 0))
    assert (maxind in [24, 25])
    return


if __name__ == "__main__":
    test_GaussRBF()
    test_LocalGaussRBF()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
from datetime import datetime

# Add ParMOO's modules to path
sys.path.insert(0, os.path.abspath('.'))
sys.path.insert(0, os.path.abspath('..'))
sys.path.insert(0, os.path.abspath('../parmoo'))


# -- Project information -----------------------------------------------------

project = 'ParMOO'
copyright = str(datetime.now().year) + ' Tyler H. Chang, Stefan M. Wild, et al.'
author = "Tyler H. Chang, Stefan M. Wild, et al."

# The full version, including alpha/beta/rc tags
exec(open("../parmoo/version.py").read())
version = __version__
release = version


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['sphinx.ext.autodoc', 'sphinx.ext.napoleon']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'python3'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# Stefan needed this
# see https://github.com/rtfd/readthedocs.org/issues/2569
master_doc = 'index'

# The latex toctree document
latex_doc = 'latex_index'

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'
html_theme_options = {'navigation_depth': 3,
                      'logo_only': True}
html_logo = './img/logo-ParMOO_white.png'

# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    'fncychap': "\\usepackage[Lenny]{fncychap}",
    'extraclassoptions': "openany",
    'preamble': r"""
    \protected\def\sphinxcrossref#1{\texttt{#1}}

    \newsavebox\mytempbox
    \definecolor{sphinxnoteBgColor}{RGB}{221,233,239}
    \renewenvironment{sphinxnote}[1]{%
    \begin{lrbox}{\mytempbox}%
    \begin{minipage}{\columnwidth}%
    \begin{sphinxlightbox}%
    \sphinxstrong{#1}}%
    {\end{sphinxlightbox}%
    \end{minipage}%
    \end{lrbox}%
    \colorbox{sphinxnoteBgColor}{\usebox{\mytempbox}}}
    """,
    }

latex_documents = [
    (latex_doc, "parmoo.tex", "ParMOO User's Manual",
     "Tyler H. Chang, Stefan M. Wild, et al.", "manual"),
     ]

latex_logo = "img/logo-ParMOO_white.png"


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []




from jax import config
from jax import jacrev
from jax import numpy as jnp
config.update("jax_enable_x64", True)


def eval_pen_jac(moop, x):
    """ Helper for testing penalty fwd/bwd evaluations """

    sx = moop._evaluate_surrogates(x)
    dsdx = jacrev(moop._evaluate_surrogates)(x)
    _, res = moop._pen_fwd(x, sx)
    dfdx = jnp.zeros((moop.o, moop.n_latent))
    for i, ei in enumerate(jnp.eye(moop.o)):
        dfdxi, dfdsi = moop._pen_bwd(res, ei)
        dfdx = dfdx.at[i].set(dfdxi + jnp.dot(dfdsi, dsdx))
    return dfdx


def eval_obj_jac(moop, x):
    """ Helper for testing objective fwd/bwd evaluations """

    sx = moop._evaluate_surrogates(x)
    dsdx = jacrev(moop._evaluate_surrogates)(x)
    _, res = moop._obj_fwd(x, sx)
    dfdx = jnp.zeros((moop.o, moop.n_latent))
    for i, ei in enumerate(jnp.eye(moop.o)):
        dfdxi, dfdsi = moop._obj_bwd(res, ei)
        dfdx = dfdx.at[i].set(dfdxi + jnp.dot(dfdsi, dsdx))
    return dfdx


def eval_con_jac(moop, x):
    """ Helper for testing constraint fwd/bwd evaluations """

    sx = moop._evaluate_surrogates(x)
    dsdx = jacrev(moop._evaluate_surrogates)(x)
    _, res = moop._con_fwd(x, sx)
    dcdx = jnp.zeros((moop.p, moop.n_latent))
    for i, ei in enumerate(jnp.eye(moop.p)):
        dcdxi, dcdsi = moop._con_bwd(res, ei)
        dcdx = dcdx.at[i].set(dcdxi + jnp.dot(dcdsi, dsdx))
    return dcdx


def test_MOOP_evaluate_penalty_grads():
    """ Check that the MOOP class handles evaluating gradients properly.

    Initialize a MOOP object and check that the evaluateGradients() function
    works correctly.

    """

    import numpy as np
    from parmoo import MOOP
    from parmoo.acquisitions import UniformWeights
    from parmoo.optimizers import GlobalSurrogate_PS
    from parmoo.searches import LatinHypercube
    from parmoo.surrogates import GaussRBF

    # Create several differentiable functions and constraints.
    def f1(x, s):
        names = ["x1", "x2", "x3"]
        return np.sum([x[i] * x[i] for i in names])

    def df1(x, s):
        return ({"x1": 2*x["x1"], "x2": 2*x["x2"], "x3": 2*x["x3"]},
                {"sim1": 0, "sim2": jnp.zeros(2)})

    def f2(x, s):
        names = ["sim1", "sim2"]
        return np.sum([jnp.dot(s[i] - 0.5, s[i] - 0.5) for i in names])

    def df2(x, s):
        return ({"x1": 0, "x2": 0, "x3": 0},
                {"sim1": 2*s["sim1"] - 1, "sim2": 2*s["sim2"] - jnp.ones(2)})

    def c1(x, s):
        return x["x1"] - 0.25

    def dc1(x, s):
        return {"x1": 1, "x2": 0, "x3": 0}, {"sim1": 0, "sim2": jnp.zeros(2)}

    def c2(x, s):
        return s["sim1"] - 0.25

    def dc2(x, s):
        return {"x1": 0, "x2": 0, "x3": 0}, {"sim1": 1, "sim2": jnp.zeros(2)}

    # Initialize a continuous MOOP with 3 variables, no sims, 1 objective
    moop1 = MOOP(GlobalSurrogate_PS)
    for i in range(3):
        moop1.addDesign({'lb': 0.0, 'ub': 1.0})
    moop1.addObjective({'obj_func': f1, 'obj_grad': df1})
    moop1.addAcquisition({'acquisition': UniformWeights})
    moop1.compile()
    # Check the shape and values of the penalty Jacobian
    assert (eval_pen_jac(moop1, np.zeros(3)).shape == (1, 3))
    assert (np.all(np.abs(eval_pen_jac(moop1, np.zeros(3))) < 1.0e-8))
    fx1 = 2.0 * np.ones((1, 3))
    assert (np.all(np.abs(eval_pen_jac(moop1, np.ones(3)) - fx1) < 1.0e-8))
    # Add a constraint and make sure that the penalty appears in the Jacobian
    moop1.addConstraint({'con_func': c1, 'con_grad': dc1})
    moop1.compile()
    assert (eval_pen_jac(moop1, np.zeros(3)).shape == (1, 3))
    assert (np.all(np.abs(eval_pen_jac(moop1, np.zeros(3))) < 1.0e-8))
    fx1[0, 0] = 3.0
    assert (np.all(np.abs(eval_pen_jac(moop1, np.ones(3)) - fx1) < 1.0e-8))
    # Create a new continuous MOOP as before but with 2 sims
    moop1 = MOOP(GlobalSurrogate_PS)
    for i in range(3):
        moop1.addDesign({'lb': 0.0, 'ub': 1.0})
    g1 = {'m': 1,
          'hyperparams': {},
          'search': LatinHypercube,
          'sim_func': lambda x: np.sqrt(sum([x[i]**2 for i in x])),
          'surrogate': GaussRBF}
    g2 = {'m': 2,
          'hyperparams': {},
          'search': LatinHypercube,
          'sim_func': lambda x: [np.sqrt(sum([(x[i] - 1.0)**2 for i in x])),
                                 np.sqrt(sum([(x[i] - 0.5)**2 for i in x]))],
          'surrogate': GaussRBF}
    moop1.addSimulation(g1, g2)
    moop1.addObjective({'obj_func': f1, 'obj_grad': df1})
    moop1.addAcquisition({'acquisition': UniformWeights})
    moop1.compile()
    # Add some data and set the surrogates
    for sn in ["sim1", "sim2"]:
        moop1.evaluateSimulation({"x1": 1, "x2": 1, "x3": 1}, sn)
    moop1._fit_surrogates()
    moop1._set_surrogate_tr(np.zeros(3), np.ones(3) * np.inf)
    # Check the Jacobian outputs with the same test cases as above
    assert (eval_pen_jac(moop1, np.zeros(3)).shape == (1, 3))
    assert (np.all(np.abs(eval_pen_jac(moop1, np.zeros(3))) < 1.0e-8))
    fx1 = 2.0 * np.ones((1, 3))
    assert (np.all(np.abs(eval_pen_jac(moop1, np.ones(3)) - fx1) < 1.0e-8))
    # Re-define the MOOP but add a constraint
    moop1 = MOOP(GlobalSurrogate_PS)
    for i in range(3):
        moop1.addDesign({'lb': 0.0, 'ub': 1.0})
    moop1.addSimulation(g1, g2)
    moop1.addObjective({'obj_func': f1, 'obj_grad': df1})
    moop1.addConstraint({'con_func': c1, 'con_grad': dc1})
    moop1.addAcquisition({'acquisition': UniformWeights})
    moop1.compile()
    for sn in ["sim1", "sim2"]:
        moop1.evaluateSimulation({"x1": 1, "x2": 1, "x3": 1}, sn)
    moop1._fit_surrogates()
    moop1._set_surrogate_tr(np.zeros(3), np.ones(3) * np.inf)
    assert (eval_pen_jac(moop1, np.zeros(3)).shape == (1, 3))
    assert (np.all(np.abs(eval_pen_jac(moop1, np.zeros(3))) < 1.0e-8))
    fx1[0, 0] = 3.0
    assert (np.all(np.abs(eval_pen_jac(moop1, np.ones(3)) - fx1) < 1.0e-8))
    # Re-define the MOOP but add objectives and constraints that use the sim
    moop1 = MOOP(GlobalSurrogate_PS)
    for i in range(3):
        moop1.addDesign({'lb': 0.0, 'ub': 1.0})
    moop1.addSimulation(g1, g2)
    moop1.addObjective({'obj_func': f1, 'obj_grad': df1})
    moop1.addObjective({'obj_func': f2, 'obj_grad': df2})
    moop1.addConstraint({'con_func': c1, 'con_grad': dc1})
    moop1.addConstraint({'con_func': c2, 'con_grad': dc2})
    moop1.addAcquisition({'acquisition': UniformWeights})
    moop1.compile()
    for sn in ["sim1", "sim2"]:
        moop1.evaluateSimulation({"x1": 1, "x2": 1, "x3": 1}, sn)
    moop1._fit_surrogates()
    moop1._set_surrogate_tr(np.zeros(3), np.ones(3) * np.inf)
    fx0 = np.zeros((2, 3))
    fx0[1, 0] = 1.0
    fx0[0, :] = 2.0
    fx0[0, 0] = 3.0
    assert (np.all(np.abs(eval_pen_jac(moop1, np.ones(3)) - fx0) < 1.0e-8))
    # Create a duplicate MOOP but adjust the design space scaling
    moop2 = MOOP(GlobalSurrogate_PS)
    moop2.addDesign({'lb': -1.0, 'ub': 1.0},
                    {'lb': 0.0, 'ub': 2.0},
                    {'lb': -0.5, 'ub': 1.5})
    moop2.addSimulation(g1, g2)
    moop2.addObjective({'obj_func': f1, 'obj_grad': df1})
    moop2.addObjective({'obj_func': f2, 'obj_grad': df2})
    moop2.addConstraint({'con_func': c1, 'con_grad': dc1})
    moop2.addConstraint({'con_func': c2, 'con_grad': dc2})
    moop2.addAcquisition({'acquisition': UniformWeights})
    moop2.compile()
    for sn in ["sim1", "sim2"]:
        moop2.evaluateSimulation({"x1": 1, "x2": 1, "x3": 1}, sn)
    moop2._fit_surrogates()
    moop2._set_surrogate_tr(np.zeros(3), np.ones(3) * np.inf)
    # After embedding inputs, the outputs should be the same for evaluations
    # at the interpolation nodes.
    x = moop1._embed({'x1': 1, 'x2': 1, 'x3': 1})
    xx = moop2._embed({'x1': 1, 'x2': 1, 'x3': 1})
    assert (np.linalg.norm(eval_pen_jac(moop1, x) -
                           eval_pen_jac(moop2, xx) * 2) < 1.e-8)
    # Now check that after compiling, jax correctly propagates pen_jac
    _, _, eval_pen1 = moop1._link()
    def pen_jac(x): return eval_pen1(x, moop1._evaluate_surrogates(x))
    moop1_pen_jac = jacrev(pen_jac)
    for xi in np.random.sample((5, 3)):
        dfdxi = moop1_pen_jac(xi)
        assert (np.all(np.abs(eval_pen_jac(moop1, xi) - dfdxi) < 1.0e-8))


def test_MOOP_evaluate_objective_grads():
    """ Check that the MOOP class handles evaluating gradients properly.

    Initialize a MOOP object and check that the evaluateGradients() function
    works correctly.

    """

    import numpy as np
    from parmoo import MOOP
    from parmoo.acquisitions import UniformWeights
    from parmoo.optimizers import GlobalSurrogate_PS
    from parmoo.searches import LatinHypercube
    from parmoo.surrogates import GaussRBF

    # Create several differentiable functions and constraints.
    def f1(x, s):
        names = ["x1", "x2", "x3"]
        return np.sum([x[i] * x[i] for i in names])

    def df1(x, s):
        return ({"x1": 2*x["x1"], "x2": 2*x["x2"], "x3": 2*x["x3"]},
                {"sim1": 0, "sim2": jnp.zeros(2)})

    def f2(x, s):
        names = ["sim1", "sim2"]
        return np.sum([jnp.dot(s[i] - 0.5, s[i] - 0.5) for i in names])

    def df2(x, s):
        return ({"x1": 0, "x2": 0, "x3": 0},
                {"sim1": 2*s["sim1"] - 1, "sim2": 2*s["sim2"] - jnp.ones(2)})

    def c1(x, s):
        return x["x1"] - 0.25

    def dc1(x, s):
        return {"x1": 1, "x2": 0, "x3": 0}, {"sim1": 0, "sim2": jnp.zeros(2)}

    def c2(x, s):
        return s["sim1"] - 0.25

    def dc2(x, s):
        return {"x1": 0, "x2": 0, "x3": 0}, {"sim1": 1, "sim2": jnp.zeros(2)}

    # Initialize a continuous MOOP with 3 variables, no sims, 1 objective
    moop1 = MOOP(GlobalSurrogate_PS)
    for i in range(3):
        moop1.addDesign({'lb': 0.0, 'ub': 1.0})
    moop1.addObjective({'obj_func': f1, 'obj_grad': df1})
    moop1.addAcquisition({'acquisition': UniformWeights})
    moop1.compile()
    # Check the shape and values of the objective Jacobian
    assert (eval_obj_jac(moop1, np.zeros(3)).shape == (1, 3))
    assert (np.all(np.abs(eval_obj_jac(moop1, np.zeros(3))) < 1.0e-8))
    fx1 = 2.0 * np.ones((1, 3))
    assert (np.all(np.abs(eval_obj_jac(moop1, np.ones(3)) - fx1) < 1.0e-8))
    # Add a constraint and make sure that the Jacobian is unchanged
    moop1.addConstraint({'con_func': c1, 'con_grad': dc1})
    moop1.compile()
    assert (eval_obj_jac(moop1, np.zeros(3)).shape == (1, 3))
    assert (np.all(np.abs(eval_obj_jac(moop1, np.zeros(3))) < 1.0e-8))
    assert (np.all(np.abs(eval_obj_jac(moop1, np.ones(3)) - fx1) < 1.0e-8))
    # Create a new continuous MOOP as before but with 2 sims
    moop1 = MOOP(GlobalSurrogate_PS)
    for i in range(3):
        moop1.addDesign({'lb': 0.0, 'ub': 1.0})
    g1 = {'m': 1,
          'hyperparams': {},
          'search': LatinHypercube,
          'sim_func': lambda x: np.sqrt(sum([x[i]**2 for i in x])),
          'surrogate': GaussRBF}
    g2 = {'m': 2,
          'hyperparams': {},
          'search': LatinHypercube,
          'sim_func': lambda x: [np.sqrt(sum([(x[i] - 1.0)**2 for i in x])),
                                 np.sqrt(sum([(x[i] - 0.5)**2 for i in x]))],
          'surrogate': GaussRBF}
    moop1.addSimulation(g1, g2)
    moop1.addObjective({'obj_func': f1, 'obj_grad': df1})
    moop1.addAcquisition({'acquisition': UniformWeights})
    moop1.compile()
    # Add some data and set the surrogates
    for sn in ["sim1", "sim2"]:
        moop1.evaluateSimulation({"x1": 1, "x2": 1, "x3": 1}, sn)
    moop1._fit_surrogates()
    moop1._set_surrogate_tr(np.zeros(3), np.ones(3) * np.inf)
    # Check the Jacobian outputs with the same test cases as above
    assert (eval_obj_jac(moop1, np.zeros(3)).shape == (1, 3))
    assert (np.all(np.abs(eval_obj_jac(moop1, np.zeros(3))) < 1.0e-8))
    fx1 = 2.0 * np.ones((1, 3))
    assert (np.all(np.abs(eval_obj_jac(moop1, np.ones(3)) - fx1) < 1.0e-8))
    # Re-define the MOOP but add a constraint
    moop1 = MOOP(GlobalSurrogate_PS)
    for i in range(3):
        moop1.addDesign({'lb': 0.0, 'ub': 1.0})
    moop1.addSimulation(g1, g2)
    moop1.addObjective({'obj_func': f1, 'obj_grad': df1})
    moop1.addConstraint({'con_func': c1, 'con_grad': dc1})
    moop1.addAcquisition({'acquisition': UniformWeights})
    moop1.compile()
    for sn in ["sim1", "sim2"]:
        moop1.evaluateSimulation({"x1": 1, "x2": 1, "x3": 1}, sn)
    moop1._fit_surrogates()
    moop1._set_surrogate_tr(np.zeros(3), np.ones(3) * np.inf)
    assert (eval_obj_jac(moop1, np.zeros(3)).shape == (1, 3))
    assert (np.all(np.abs(eval_obj_jac(moop1, np.zeros(3))) < 1.0e-8))
    assert (np.all(np.abs(eval_obj_jac(moop1, np.ones(3)) - fx1) < 1.0e-8))
    # Re-define the MOOP but add objectives and constraints that use the sim
    moop1 = MOOP(GlobalSurrogate_PS)
    for i in range(3):
        moop1.addDesign({'lb': 0.0, 'ub': 1.0})
    moop1.addSimulation(g1, g2)
    moop1.addObjective({'obj_func': f1, 'obj_grad': df1})
    moop1.addObjective({'obj_func': f2, 'obj_grad': df2})
    moop1.addConstraint({'con_func': c1, 'con_grad': dc1})
    moop1.addConstraint({'con_func': c2, 'con_grad': dc2})
    moop1.addAcquisition({'acquisition': UniformWeights})
    moop1.compile()
    for sn in ["sim1", "sim2"]:
        moop1.evaluateSimulation({"x1": 1, "x2": 1, "x3": 1}, sn)
    moop1._fit_surrogates()
    moop1._set_surrogate_tr(np.zeros(3), np.ones(3) * np.inf)
    fx0 = np.zeros((2, 3))
    fx0[0, :] = 2.0
    assert (np.all(np.abs(eval_obj_jac(moop1, np.ones(3)) - fx0) < 1.0e-8))
    # Create a duplicate MOOP but adjust the design space scaling
    moop2 = MOOP(GlobalSurrogate_PS)
    moop2.addDesign({'lb': -1.0, 'ub': 1.0},
                    {'lb': 0.0, 'ub': 2.0},
                    {'lb': -0.5, 'ub': 1.5})
    moop2.addSimulation(g1, g2)
    moop2.addObjective({'obj_func': f1, 'obj_grad': df1})
    moop2.addObjective({'obj_func': f2, 'obj_grad': df2})
    moop2.addConstraint({'con_func': c1, 'con_grad': dc1})
    moop2.addConstraint({'con_func': c2, 'con_grad': dc2})
    moop2.addAcquisition({'acquisition': UniformWeights})
    moop2.compile()
    for sn in ["sim1", "sim2"]:
        moop2.evaluateSimulation({"x1": 1, "x2": 1, "x3": 1}, sn)
    moop2._fit_surrogates()
    moop2._set_surrogate_tr(np.zeros(3), np.ones(3) * np.inf)
    # After embedding inputs, the outputs should be the same for evaluations
    # at the interpolation nodes.
    x = moop1._embed({'x1': 1, 'x2': 1, 'x3': 1})
    xx = moop2._embed({'x1': 1, 'x2': 1, 'x3': 1})
    assert (np.linalg.norm(eval_obj_jac(moop1, x) -
                           eval_obj_jac(moop2, xx) * 2) < 1.e-8)
    eval_obj1, _, _ = moop1._link()
    # Now check that after compiling, jax correctly propagates obj_jac
    def obj_jac(x): return eval_obj1(x, moop1._evaluate_surrogates(x))
    moop1_obj_jac = jacrev(obj_jac)
    for xi in np.random.sample((5, 3)):
        dfdxi = moop1_obj_jac(xi)
        assert (np.all(np.abs(eval_obj_jac(moop1, xi) - dfdxi) < 1.0e-8))


def test_MOOP_evaluate_constraint_grads():
    """ Check that the MOOP class handles evaluating gradients properly.

    Initialize a MOOP object and check that the evaluateGradients() function
    works correctly.

    """

    import numpy as np
    from parmoo import MOOP
    from parmoo.acquisitions import UniformWeights
    from parmoo.optimizers import GlobalSurrogate_PS
    from parmoo.searches import LatinHypercube
    from parmoo.surrogates import GaussRBF

    # Create several differentiable functions and constraints.
    def f1(x, s):
        names = ["x1", "x2", "x3"]
        return np.sum([x[i] * x[i] for i in names])

    def df1(x, s):
        return ({"x1": 2*x["x1"], "x2": 2*x["x2"], "x3": 2*x["x3"]},
                {"sim1": 0, "sim2": jnp.zeros(2)})

    def f2(x, s):
        names = ["sim1", "sim2"]
        return np.sum([jnp.dot(s[i] - 0.5, s[i] - 0.5) for i in names])

    def df2(x, s):
        return ({"x1": 0, "x2": 0, "x3": 0},
                {"sim1": 2*s["sim1"] - 1, "sim2": 2*s["sim2"] - jnp.ones(2)})

    def c1(x, s):
        return x["x1"] - 0.25

    def dc1(x, s):
        return {"x1": 1, "x2": 0, "x3": 0}, {"sim1": 0, "sim2": jnp.zeros(2)}

    def c2(x, s):
        return s["sim1"] - 0.25

    def dc2(x, s):
        return {"x1": 0, "x2": 0, "x3": 0}, {"sim1": 1, "sim2": jnp.zeros(2)}

    # Initialize a continuous MOOP with 3 variables, no sims, 1 objective
    moop1 = MOOP(GlobalSurrogate_PS)
    for i in range(3):
        moop1.addDesign({'lb': 0.0, 'ub': 1.0})
    moop1.addObjective({'obj_func': f1, 'obj_grad': df1})
    moop1.addAcquisition({'acquisition': UniformWeights})
    moop1.compile()
    # Check the shape and values of the constraint Jacobian
    assert (eval_con_jac(moop1, np.zeros(3)).size == 0)
    # Add a constraint and make sure it appears in the Jacobian
    moop1.addConstraint({'con_func': c1, 'con_grad': dc1})
    moop1.compile()
    assert (eval_con_jac(moop1, np.zeros(3)).shape == (1, 3))
    cx1 = np.zeros((1, 3))
    cx1[0, 0] = 1.0
    assert (np.all(np.abs(eval_con_jac(moop1, np.zeros(3)) - cx1) < 1.0e-8))
    assert (np.all(np.abs(eval_con_jac(moop1, np.ones(3)) - cx1) < 1.0e-8))
    # Create a new continuous MOOP as before but with 2 sims
    moop1 = MOOP(GlobalSurrogate_PS)
    for i in range(3):
        moop1.addDesign({'lb': 0.0, 'ub': 1.0})
    g1 = {'m': 1,
          'hyperparams': {},
          'search': LatinHypercube,
          'sim_func': lambda x: np.sqrt(sum([x[i]**2 for i in x])),
          'surrogate': GaussRBF}
    g2 = {'m': 2,
          'hyperparams': {},
          'search': LatinHypercube,
          'sim_func': lambda x: [np.sqrt(sum([(x[i] - 1.0)**2 for i in x])),
                                 np.sqrt(sum([(x[i] - 0.5)**2 for i in x]))],
          'surrogate': GaussRBF}
    moop1.addSimulation(g1, g2)
    moop1.addObjective({'obj_func': f1, 'obj_grad': df1})
    moop1.addConstraint({'con_func': c1, 'con_grad': dc1})
    moop1.addAcquisition({'acquisition': UniformWeights})
    moop1.compile()
    for sn in ["sim1", "sim2"]:
        moop1.evaluateSimulation({"x1": 1, "x2": 1, "x3": 1}, sn)
    moop1._fit_surrogates()
    moop1._set_surrogate_tr(np.zeros(3), np.ones(3) * np.inf)
    assert (eval_con_jac(moop1, np.zeros(3)).shape == (1, 3))
    assert (np.all(np.abs(eval_con_jac(moop1, np.zeros(3)) - cx1) < 1.0e-8))
    assert (np.all(np.abs(eval_con_jac(moop1, np.ones(3)) - cx1) < 1.0e-8))
    # Re-define the MOOP but add objectives and constraints that use the sim
    moop1 = MOOP(GlobalSurrogate_PS)
    for i in range(3):
        moop1.addDesign({'lb': 0.0, 'ub': 1.0})
    moop1.addSimulation(g1, g2)
    moop1.addObjective({'obj_func': f1, 'obj_grad': df1})
    moop1.addObjective({'obj_func': f2, 'obj_grad': df2})
    moop1.addConstraint({'con_func': c1, 'con_grad': dc1})
    moop1.addConstraint({'con_func': c2, 'con_grad': dc2})
    moop1.addAcquisition({'acquisition': UniformWeights})
    moop1.compile()
    for sn in ["sim1", "sim2"]:
        moop1.evaluateSimulation({"x1": 1, "x2": 1, "x3": 1}, sn)
    moop1._fit_surrogates()
    moop1._set_surrogate_tr(np.zeros(3), np.ones(3) * np.inf)
    cx2 = np.zeros((2, 3))
    cx2[0, 0] = 1.0
    assert (np.all(np.abs(eval_con_jac(moop1, np.ones(3)) - cx2) < 1.0e-8))
    # Create a duplicate MOOP but adjust the design space scaling
    moop2 = MOOP(GlobalSurrogate_PS)
    moop2.addDesign({'lb': -1.0, 'ub': 1.0},
                    {'lb': 0.0, 'ub': 2.0},
                    {'lb': -0.5, 'ub': 1.5})
    moop2.addSimulation(g1, g2)
    moop2.addObjective({'obj_func': f1, 'obj_grad': df1})
    moop2.addObjective({'obj_func': f2, 'obj_grad': df2})
    moop2.addConstraint({'con_func': c1, 'con_grad': dc1})
    moop2.addConstraint({'con_func': c2, 'con_grad': dc2})
    moop2.addAcquisition({'acquisition': UniformWeights})
    moop2.compile()
    for sn in ["sim1", "sim2"]:
        moop2.evaluateSimulation({"x1": 1, "x2": 1, "x3": 1}, sn)
    moop2._fit_surrogates()
    moop2._set_surrogate_tr(np.zeros(3), np.ones(3) * np.inf)
    # After embedding inputs, the outputs should be the same for evaluations
    # at the interpolation nodes.
    x = moop1._embed({'x1': 1, 'x2': 1, 'x3': 1})
    xx = moop2._embed({'x1': 1, 'x2': 1, 'x3': 1})
    assert (np.linalg.norm(eval_con_jac(moop1, x) -
                           eval_con_jac(moop2, xx) * 2) < 1.e-8)
    _, eval_con1, _ = moop1._link()
    # Now check that after compiling, jax correctly propagates con_jac
    def con_jac(x): return eval_con1(x, moop1._evaluate_surrogates(x))
    moop1_con_jac = jacrev(con_jac)
    for xi in np.random.sample((5, 3)):
        dfdxi = moop1_con_jac(xi)
        assert (np.all(np.abs(eval_con_jac(moop1, xi) - dfdxi) < 1.0e-8))


def test_MOOP_solve_with_grads():
    """ Check that the MOOP class propagates gradients correctly to solvers.

    Initialize a simple convex MOOP and check that the gradient-based solver
    matches the gradient-free solver's solutions.

    """

    import numpy as np
    from parmoo import MOOP
    from parmoo.acquisitions import FixedWeights
    from parmoo.optimizers import GlobalSurrogate_BFGS, GlobalSurrogate_PS
    from parmoo.searches import LatinHypercube
    from parmoo.surrogates import GaussRBF

    # Create several differentiable simulation groups
    g1 = {'m': 1,
          'search': LatinHypercube,
          'sim_func': lambda x: [np.sqrt(sum([x[i]**2 for i in x]))],
          'surrogate': GaussRBF,
          'hyperparams': {'search_budget': 100}}
    g2 = {'m': 2,
          'search': LatinHypercube,
          'sim_func': lambda x: [np.sqrt(sum([(x[i]-0.5)**2 for i in x])),
                                 np.sqrt(sum([(x[i]-1.0)**2 for i in x]))],
          'surrogate': GaussRBF,
          'hyperparams': {'search_budget': 100}}

    # Create several differentiable functions and constraints
    def f1(x, s):
        return x["x1"] ** 2

    def df1(x, s):
        return ({"x1": 2*x["x1"]},
                {"sim1": 0, "sim2": jnp.zeros(2)})

    def f2(x, s):
        names = ["sim1", "sim2"]
        return np.sum([jnp.dot(s[i] - 0.5, s[i] - 0.5) for i in names])

    def df2(x, s):
        return ({"x1": 0},
                {"sim1": 2*s["sim1"] - 1, "sim2": 2*s["sim2"] - jnp.ones(2)})

    def c1(x, s):
        return x["x1"] - 0.25

    def dc1(x, s):
        return {"x1": 1}, {"sim1": 0, "sim2": jnp.zeros(2)}

    def c2(x, s):
        return s["sim1"] - 0.25

    def dc2(x, s):
        return {"x1": 0}, {"sim1": 1, "sim2": jnp.zeros(2)}

    # Initialize 2 continuous MOOPs with 1 design var, 2 sims, and 3 objs
    moop1 = MOOP(GlobalSurrogate_BFGS, hyperparams={'opt_restarts': 2,
                                                    'np_random_gen': 0})
    moop1.addDesign({'lb': 0.0, 'ub': 1.0})
    moop1.addSimulation(g1, g2)
    moop1.addObjective({'obj_func': f1, 'obj_grad': df1})
    moop1.addObjective({'obj_func': f2, 'obj_grad': df2})
    moop1.addConstraint({'constraint': c1, 'con_grad': dc1})
    moop1.addConstraint({'constraint': c2, 'con_grad': dc2})
    moop1.addAcquisition({'acquisition': FixedWeights,
                          'hyperparams': {'weights': np.ones(2) / 2}})
    moop1.solve(0)
    moop2 = MOOP(GlobalSurrogate_PS, hyperparams={'np_random_gen': 0})
    moop2.addDesign({'lb': 0.0, 'ub': 1.0})
    moop2.addSimulation(g1, g2)
    moop2.addObjective({'obj_func': f1})
    moop2.addObjective({'obj_func': f2})
    moop2.addConstraint({'constraint': c1})
    moop2.addConstraint({'constraint': c2})
    moop2.addAcquisition({'acquisition': FixedWeights,
                          'hyperparams': {'weights': np.ones(2) / 2}})
    moop2.solve(0)
    b1 = moop1.iterate(1)
    b2 = moop2.iterate(1)
    # Check that same solutions were found
    for x1, x2 in zip(b1, b2):
        assert (np.abs(x1[0]["x1"] - x2[0]["x1"]) < 0.1)


if __name__ == "__main__":
    test_MOOP_evaluate_penalty_grads()
    test_MOOP_evaluate_objective_grads()
    test_MOOP_evaluate_constraint_grads()
    test_MOOP_solve_with_grads()




""" Use ParMOO to solve the DTLZ3 problem, treating DTLZ3 as an objective.

Uses unnamed variables, the g1_sim simulation function, and the
dtlz3_obj objective functions to define the problem.

"""

import numpy as np
from parmoo import MOOP
from parmoo.optimizers import GlobalSurrogate_PS
from parmoo.surrogates import GaussRBF
from parmoo.acquisitions import UniformWeights
from parmoo.searches import LatinHypercube
from parmoo.simulations.dtlz import g1_sim as sim_func
from parmoo.objectives.dtlz import dtlz3_obj as obj_func

# Set the problem dimensions
NUM_DES = 3
NUM_OBJ = 3

# Create a MOOP
moop = MOOP(GlobalSurrogate_PS)

# Add NUM_DES continuous design variables
for i in range(NUM_DES):
    moop.addDesign({'ub': 1.0, 'lb': 0.0,
                    'des_type': "continuous", 'des_tol': 1.0e-8})

# Add the simulation
moop.addSimulation({'name': "g1",
                    'm': 1,
                    'sim_func': sim_func(moop.getDesignType(),
                                         num_obj=NUM_OBJ,
                                         offset=0.6),
                    'search': LatinHypercube,
                    'surrogate': GaussRBF,
                    'hyperparams': {}})

# Add NUM_OBJ objective functions
for i in range(NUM_OBJ):
    moop.addObjective({'name': f"DTLZ3 objective {i+1}",
                       'obj_func': obj_func(moop.getDesignType(),
                                            moop.getSimulationType(),
                                            obj_ind=i, num_obj=NUM_OBJ)})

# Define 2 constraints to nudge the solver in the right direction

def min_constraint(x, sx):
    """ x[NUM_OBJ-1:NUM_DES] >= 0.5 """

    fx = 0.0
    for i in range(NUM_OBJ - 1, NUM_DES):
        fx += (0.5 - x[f"x{i+1}"])
    return fx

def max_constraint(x, sx):
    """ x[NUM_OBJ-1:NUM_DES] <= 0.7 """

    fx = 0.0
    for i in range(NUM_OBJ - 1, NUM_DES):
        fx += (x[f"x{i+1}"] - 0.7)
    return fx

# Add 2 constraints to the problem
moop.addConstraint({'name': "Lower Bounds", 'con_func': min_constraint})
moop.addConstraint({'name': "Upper Bounds", 'con_func': max_constraint})

# Add 5 acquisition functions
for i in range(5):
    moop.addAcquisition({'acquisition': UniformWeights, 'hyperparams': {}})

# Solve the problem with 5 iterations
moop.solve(5)

# Check that 125 simulations were evaluated and solutions were found
assert(moop.getObjectiveData()['x1'].shape[0] == 125)
assert(moop.getSimulationData()['g1'].shape[0] == 125)
assert(moop.getPF().shape[0] > 0)




""" Implementations of the GlobalSearch class.

This module contains implementations of the GlobalSearch ABC, which are based
on the Latin hypercube design.

The classes include:
 * ``LatinHypercube`` -- Latin hypercube sampling

"""

import numpy as np
from parmoo.structs import GlobalSearch
from parmoo.util import xerror
from scipy.stats import qmc


class LatinHypercube(GlobalSearch):
    """ Implementation of a Latin hypercube search.

    This GlobalSearch strategy uses a Latin hypercube design to sample in the
    design space.

    """

    # Slots for the LatinHypercube class
    __slots__ = ['lb', 'ub', 'budget', 'sampler', 'np_rng']

    def __init__(self, m, lb, ub, hyperparams):
        """ Constructor for the LatinHypercube GlobalSearch class.

        Args:
            m (int): The number of simulation outputs (unused by this class).

            lb (numpy.ndarray): A 1d array of lower bounds for the design
                region. The number of design variables is inferred from the
                dimension of lb.

            ub (numpy.ndarray): A 1d array of upper bounds for the design
                region. The dimension must match ub.

            hyperparams (dict): A dictionary of hyperparameters for the
                LatinHypercube design. It may contain:
                 * search_budget (int): The sim eval budget for the search

        Returns:
            LatinHypercube: A new LatinHypercube object.

        """

        # Check inputs
        xerror(o=m, lb=lb, ub=ub, hyperparams=hyperparams)
        self.n = lb.size
        # Assign the bounds
        self.lb = lb
        self.ub = ub
        # Check for a search budget
        if 'search_budget' in hyperparams:
            if isinstance(hyperparams['search_budget'], int):
                if hyperparams['search_budget'] < 0:
                    raise ValueError("hyperparams['search_budget'] must "
                                     "be nonnegative")
                else:
                    self.budget = hyperparams['search_budget']
            else:
                raise ValueError("hyperparams['search_budget'] must "
                                 "be an int")
        else:
            self.budget = 100
        # Check the hyperparameter dictionary for random generator
        if 'np_random_gen' in hyperparams:
            if isinstance(hyperparams['np_random_gen'], np.random.Generator):
                self.np_rng = hyperparams['np_random_gen']
            else:
                raise TypeError("When present, hyperparams['np_random_gen'] "
                                "must be an instance of the class "
                                "numpy.random.Generator")
        else:
            self.np_rng = np.random.default_rng()
        self.sampler = qmc.LatinHypercube(d=self.n, seed=self.np_rng)
        return

    def startSearch(self, lb, ub):
        """ Begin a new Latin hypercube sampling.

        Args:
            lb (numpy.ndarray): A 1d array of lower bounds for the design
                region. The dimension must match n.

            ub (numpy.ndarray): A 1d array of upper bounds for the design
                region. The dimension must match n.

        Returns:
            np.ndarray: A 2d array, containing the list of design points
            to be evaluated.

        """

        # Check inputs
        xerror(lb=lb, ub=ub)
        # Assign the bounds
        self.lb = lb
        self.ub = ub
        # If the budget is 0, just return an empty array
        if self.budget == 0:
            return np.asarray([])
        # Otherwise, return a n-dimensional Latin hypercube design
        else:
            return qmc.scale(self.sampler.random(n=self.budget),
                             self.lb, self.ub)

    def resumeSearch(self):
        """ Resume a previous Latin hypercube sampling.

        Returns:
            np.ndarray: A 2d array, containing the list of design points
            to be evaluated.

        """

        # If the budget is 0, just return an empty array
        if self.budget == 0:
            return np.asarray([])
        # Otherwise, return a n-dimensional Latin hypercube design
        else:
            return qmc.scale(self.sampler.random(n=self.budget),
                             self.lb, self.ub)




""" Use ParMOO to solve a convex, user-defined problem.

Uses unnamed variables and public function definitions to define the problem.

Also turns on checkpointing, to test ParMOO's checkpointing of unnamed
variables and globally defined functions.

"""

import numpy as np
import os
from parmoo import MOOP
from parmoo.optimizers import LocalSurrogate_BFGS
from parmoo.surrogates import GaussRBF
from parmoo.acquisitions import RandomConstraint
from parmoo.searches import LatinHypercube

# Set the problem dimensions
NUM_DES = 3

# For this test, use all user-defined functions

def sim(x):
    " User sim for sample problem. "
    return [sum([x[i] for i in x])]

def obj1_func(x, sx):
    " User obj1 for sample problem. "
    return sum([sx[i] for i in sx])

def obj1_grad(x, sx):
    " User obj1 for sample problem. "
    dx, ds = {}, {}
    for key in x:
        dx[key] = 0.
    for key in sx:
        ds[key] = 1.
    return dx, ds

def obj2_func(x, sx):
    " User obj2 for sample problem. "
    return 1.0 - sum([sx[i] for i in sx])

def obj2_grad(x, sx):
    " User obj2 for sample problem. "
    dx, ds = {}, {}
    for key in x:
        dx[key] = 0.
    for key in sx:
        ds[key] = -1.
    return dx, ds

# Create a MOOP
moop = MOOP(LocalSurrogate_BFGS)

# Add NUM_DES continuous design variables
for i in range(NUM_DES):
    moop.addDesign({'ub': 1.0, 'lb': 0.0,
                    'des_type': "continuous", 'des_tol': 1.0e-8})

# Add the simulation
moop.addSimulation({'name': "Sample sim",
                    'm': 1,
                    'sim_func': sim,
                    'search': LatinHypercube,
                    'surrogate': GaussRBF,
                    'hyperparams': {}})

# Add user objective functions
moop.addObjective({'obj_func': obj1_func, 'obj_grad': obj1_grad},
                  {'obj_func': obj2_func, 'obj_grad': obj2_grad})

# Add 10 acquisition functions
for i in range(10):
    moop.addAcquisition({'acquisition': RandomConstraint, 'hyperparams': {}})

# Solve the problem with 5 iterations with checkpointing on
moop.setCheckpoint(True)
moop.solve(5)

# Check that 150 simulations were evaluated and solutions are feasible
assert(moop.getObjectiveData().shape[0] == 150)
assert(moop.getSimulationData()['Sample sim'].shape[0] == 150)
assert(moop.getPF().shape[0] > 0)

# Clean up test directory (remove checkpoint files)
os.remove("parmoo.moop")
os.remove("parmoo.simdb.json")
os.remove("parmoo.surrogate.1")
os.remove("parmoo.optimizer")




import numpy as np
from parmoo import MOOP
from parmoo.searches import LatinHypercube
from parmoo.surrogates import GaussRBF
from parmoo.acquisitions import UniformWeights
from parmoo.optimizers import GlobalSurrogate_PS
import logging

# Create a new MOOP -- fix the random seed with the hyperparams
my_moop = MOOP(GlobalSurrogate_PS, hyperparams={'np_random_gen': 0})

# Add 1 continuous and 1 categorical design variable
my_moop.addDesign({'name': "x1",
                   'des_type': "continuous",
                   'lb': 0.0, 'ub': 1.0})
my_moop.addDesign({'name': "x2", 'des_type': "categorical",
                   'levels': 3})

# Create a simulation function
def sim_func(x):
   if x["x2"] == 0:
      return np.array([(x["x1"] - 0.2) ** 2, (x["x1"] - 0.8) ** 2])
   else:
      return np.array([99.9, 99.9])

# Add the simulation function to the MOOP
my_moop.addSimulation({'name': "MySim",
                       'm': 2,
                       'sim_func': sim_func,
                       'search': LatinHypercube,
                       'surrogate': GaussRBF,
                       'hyperparams': {'search_budget': 20}})

# Define the 2 objectives as named Python functions
def obj1(x, s): return s["MySim"][0]
def obj2(x, s): return s["MySim"][1]

# Define the constraint as a function
def const(x, s): return 0.1 - x["x1"]

# Add 2 objectives
my_moop.addObjective({'name': "f1", 'obj_func': obj1})
my_moop.addObjective({'name': "f2", 'obj_func': obj2})

# Add 1 constraint
my_moop.addConstraint({'name': "c1", 'constraint': const})

# Add 3 acquisition functions (generates batches of size 3)
for i in range(3):
   my_moop.addAcquisition({'acquisition': UniformWeights,
                           'hyperparams': {}})

# Turn on logging with timestamps
logging.basicConfig(level=logging.INFO,
                    format='%(asctime)s %(levelname)-8s %(message)s',
                    datefmt='%Y-%m-%d %H:%M:%S')

# Use checkpointing without saving a separate data file (in "parmoo.moop" file)
my_moop.setCheckpoint(True, checkpoint_data=False, filename="parmoo")

# Solve the problem with 4 iterations
my_moop.solve(4)

# Create a new MOOP object and reload the MOOP from parmoo.moop file
new_moop = MOOP(GlobalSurrogate_PS)
new_moop.load("parmoo")

# Do another iteration
new_moop.solve(5)

# Display the solution
results = new_moop.getPF()
print(results, "\n dtype=" + str(results.dtype))




def test_AcquisitionFunction():
    """ Test that the AcquisitionFunction ABC raises a TypeError"""

    from parmoo.structs import AcquisitionFunction
    import pytest

    with pytest.raises(TypeError):
        AcquisitionFunction(0, 0, 0, 0)


def test_GlobalSearch():
    """ Test that the GlobalSearch ABC raises a TypeError"""

    from parmoo.structs import GlobalSearch
    import pytest

    with pytest.raises(TypeError):
        GlobalSearch(0, 0, 0, 0)


def test_SurrogateFunction():
    """ Test that the SurrogateFunction ABC raises a TypeError"""

    from parmoo.structs import SurrogateFunction
    import pytest

    with pytest.raises(TypeError):
        SurrogateFunction(0, 0, 0, 0)


def test_SurrogateOptimizer():
    """ Test that the SurrogateFunction ABC raises a TypeError"""

    from parmoo.structs import SurrogateOptimizer
    import pytest

    with pytest.raises(TypeError):
        SurrogateOptimizer(0, 0, 0, 0)


if __name__ == "__main__":
    test_AcquisitionFunction()
    test_GlobalSearch()
    test_SurrogateFunction()
    test_SurrogateOptimizer()



# """
# Apply the libE generator to dtlz2.
#
# Execute via one of the following commands (e.g. 3 workers):
#    mpiexec -np 4 python3 test_6-hump_camel_persistent_uniform_sampling.py
#    python3 test_6-hump_camel_persistent_uniform_sampling.py \
#           --nworkers 3 --comms local
#    python3 test_6-hump_camel_persistent_uniform_sampling.py \
#           --nworkers 3 --comms tcp
#
# The number of concurrent evaluations of the objective function will be 4-1=3.
#
# """

""" Test the parmoo_gen generator function for libE.

Create a libEnsemble run and test the parmoo generator.

"""

import numpy as np
from parmoo.extras.libe import libE_MOOP
from parmoo.optimizers import GlobalSurrogate_PS
from parmoo.surrogates import GaussRBF
from parmoo.searches import LatinHypercube
from parmoo.acquisitions import RandomConstraint
from parmoo.tests.libe_tests.libe_funcs import *

# On MacOS and Windows, libE runs using Python MP must be
# enclosed in an "if __name__ == '__main__':" block, as below
if __name__ == "__main__":
    # Create a libE_MOOP
    moop = libE_MOOP(GlobalSurrogate_PS, hyperparams={'sim_dirs_make': False})
    # Add n design vars
    for i in range(n):
        moop.addDesign({'lb': 0.0, 'ub': 1.0})
    # Add simulation
    moop.addSimulation({'name': "DTLZ2",
                        'm': o,
                        'sim_func': dtlz2_sim_named,
                        'hyperparams': {'search_budget': 20},
                        'search': LatinHypercube,
                        'surrogate': GaussRBF,
                        'des_tol': 0.00000001})
    # Add o objectives
    moop.addObjective({'obj_func': obj1_named})
    moop.addObjective({'obj_func': obj2_named})
    moop.addObjective({'obj_func': obj3_named})
    # Add 4 acquisition functions
    for i in range(4):
        moop.addAcquisition({'acquisition': RandomConstraint})
    moop.compile()
    # Solve with 10 additional iterations
    moop.solve(sim_max=60)
    assert(moop.getObjectiveData()['x1'].shape[0] == 60)




""" Use ParMOO to "solve" an infeasible problem.

Ensure that ParMOO correctly identifies the most nearly feasible point.

"""

from parmoo import MOOP
from parmoo.optimizers import GlobalSurrogate_BFGS
from parmoo.surrogates import GaussRBF
from parmoo.acquisitions import UniformWeights
from parmoo.searches import LatinHypercube
import os
import numpy as np

# For this test, use all user-defined functions

def sim(x):
    " User sim for sample problem. "
    return [sum([x[name] for name in x])]

def obj_func(x, sx):
    " User objective for sample problem. "
    return 3.0 * sx["Sample sim"]

def obj_grad(x, sx):
    " User gradient for sample problem. "
    dx, ds = {}, {}
    for key in x:
        dx[key] = 0.
    for key in sx:
        ds[key] = 0.
    ds["Sample sim"] = 3.0
    return dx, ds

def const_func(x, sx):
    return sum([(x[name] - 1.0) ** 2 for name in x])

def const_grad(x, sx):
    " User constraint gradient for sample problem. "
    dx, ds = {}, {}
    for key in x:
        dx[key] = -2.
    for key in sx:
        ds[key] = 0.
    return dx, ds

# Create a MOOP
moop = MOOP(GlobalSurrogate_BFGS)

# Add 2 continuous design variables
for i in range(2):
    moop.addDesign({'name': f"x{i+1}", 'ub': 1.0, 'lb': 0.0,
                    'des_type': "continuous", 'des_tol': 1.0e-8})

# Add the simulation
moop.addSimulation({'name': "Sample sim",
                    'm': 1,
                    'sim_func': sim,
                    'search': LatinHypercube,
                    'surrogate': GaussRBF,
                    'hyperparams': {}})

# Add user objective functions
moop.addObjective({'name': "My objective",
                   'obj_func': obj_func,
                   'obj_grad': obj_grad})

# Add user constraint functions, with only 1 feasible point
moop.addConstraint({'name': "My constraint",
                    'con_func': const_func,
                    'con_grad': const_grad})

# Add 1 acquisition function
moop.addAcquisition({'acquisition': UniformWeights, 'hyperparams': {}})

# Solve the problem with 5 iterations to search for feasible point
moop.solve(5)

# Check that all solutions are (nearly) feasible
assert(all([np.all(np.abs(moop.getPF()[name] - 1.0 < 1.0e-8))
            for name in moop.getDesignType().names]))




import numpy as np
import pandas as pd
from parmoo import MOOP
from parmoo.searches import LatinHypercube
from parmoo.surrogates import GaussRBF
from parmoo.acquisitions import RandomConstraint, FixedWeights
from parmoo.optimizers import LocalSurrogate_BFGS
from parmoo.objectives import SingleSimObjective, SingleSimGradient
import logging

# Switch to using the LocalSurrogate_BFGS solver to solve surrogate problems
# in a trust region with multistart LBFGSB
my_moop = MOOP(LocalSurrogate_BFGS, hyperparams={'np_random_gen': 0})

# Massive 50-variable black-box optimization problem
# Completely hopeless for methods that rely on global models
for i in range(1, 51):
    my_moop.addDesign({'name': f"x{i}",
                       'des_type': "continuous",
                       'lb': 0.0, 'ub': 1.0})

# A simple convex simulation output with 2 outputs
def sim_func(x):
    xx = np.zeros(50)
    for i in range(50):
        xx[i] = x[f"x{i+1}"]
    # 25 variables that don't affect tradeoff, but need to be minimized
    tail = np.linalg.norm(xx[25:] - 0.5) ** 2 / 25
    # 25 variables that do affect tradeoff
    s1 = np.linalg.norm(xx[:25] - 0.2) ** 2 / 25 + tail
    s2 = np.linalg.norm(xx[:25] - 0.8) ** 2 / 25 + tail
    return np.array([s1, s2])

# Using a local surrogate to dodge the curse of dimensionality
# Notice that search_budget has to be greater than the number of variables
my_moop.addSimulation({'name': "MySim",
                       'm': 2,
                       'sim_func': sim_func,
                       'search': LatinHypercube,
                       'surrogate': GaussRBF,
                       'hyperparams': {'search_budget': 200}})

# 2 objectives (using the SingleSimObjective library objective to minimize a
# single output of the simulation function)
my_moop.addObjective({'name': "f1",
                      'obj_func':
                      SingleSimObjective(my_moop.getDesignType(),
                                         my_moop.getSimulationType(),
                                         ("MySim", 0)),
                      'obj_grad':
                      SingleSimGradient(my_moop.getDesignType(),
                                        my_moop.getSimulationType(),
                                        ("MySim", 0))})
my_moop.addObjective({'name': "f2",
                      'obj_func':
                      SingleSimObjective(my_moop.getDesignType(),
                                         my_moop.getSimulationType(),
                                         ("MySim", 1)),
                      'obj_grad':
                      SingleSimGradient(my_moop.getDesignType(),
                                        my_moop.getSimulationType(),
                                        ("MySim", 1))})

# When solving big problems, it's often better to fix some acquisitions so
# we can focus on a few high-quality solutions
my_moop.addAcquisition({'acquisition': FixedWeights,
                        'hyperparams': {'weights': np.eye(2)[0]}})
my_moop.addAcquisition({'acquisition': FixedWeights,
                        'hyperparams': {'weights': np.eye(2)[1]}})
my_moop.addAcquisition({'acquisition': FixedWeights,
                        'hyperparams': {'weights': np.ones(2) / 2}})
# Keep one randomized acquisition to get some coverage of the Pareto front
my_moop.addAcquisition({'acquisition': RandomConstraint,
                        'hyperparams': {}})

# Turn on logging with timestamps
logging.basicConfig(level=logging.INFO,
                    format='%(asctime)s %(levelname)-8s %(message)s',
                    datefmt='%Y-%m-%d %H:%M:%S')

# 50 iterations * 4 acquisition funcs + 50 point search = 250 eval budget
# This could take a few minutes to run...
my_moop.solve(50)

# Display the values of x26, ..., x50 for all solution points
results = my_moop.getPF(format="pandas")
results[[f"x{i}" for i in range(26, 51)]].to_csv("local_method.csv")

# Plot results -- must have extra viz dependencies installed
from parmoo.viz import scatter
# The optional arg `output` exports directly to jpeg instead of interactive mode
scatter(my_moop, output="jpeg")




""" Use ParMOO to solve the DTLZ1 problem, treating DTLZ1 as a simulation.

Uses named variables, the dtlz1_sim simulation function, and the
single_sim_out objective functions to define the problem.

"""

import numpy as np
from parmoo import MOOP
from parmoo.optimizers import GlobalSurrogate_BFGS
from parmoo.surrogates import GaussRBF
from parmoo.acquisitions import RandomConstraint
from parmoo.searches import LatinHypercube
from parmoo.simulations.dtlz import dtlz1_sim as sim_func
from parmoo.objectives import SingleSimObjective as obj_func
from parmoo.objectives import SingleSimGradient as obj_grad

# Set the problem dimensions
NUM_DES = 3
NUM_OBJ = 3

# Create a MOOP
moop = MOOP(GlobalSurrogate_BFGS)

# Add NUM_DES continuous design variables
for i in range(NUM_DES):
    moop.addDesign({'name': f"x{i+1}", 'ub': 1.0, 'lb': 0.0,
                    'des_type': "continuous", 'des_tol': 1.0e-8})

# Add the simulation
moop.addSimulation({'name': "DTLZ1",
                    'm': NUM_OBJ,
                    'sim_func': sim_func(moop.getDesignType(),
                                         num_obj=NUM_OBJ,
                                         offset=0.6),
                    'search': LatinHypercube,
                    'surrogate': GaussRBF,
                    'hyperparams': {}})

# Add NUM_OBJ objective functions
for i in range(NUM_OBJ):
    moop.addObjective({'name': f"f{i+1}",
                       'obj_func': obj_func(moop.getDesignType(),
                                            moop.getSimulationType(),
                                            ("DTLZ1", i), goal="min"),
                       'obj_grad': obj_grad(moop.getDesignType(),
                                            moop.getSimulationType(),
                                            ("DTLZ1", i), goal="min"),
                      })

# Define 2 constraints to nudge the solver in the right direction

def min_constraint_func(x, sx):
    """ x[NUM_OBJ-1:NUM_DES] >= 0.55 """

    fx = 0.0
    for i in range(NUM_OBJ - 1, NUM_DES):
        fx += (0.55 - x[f"x{i+1}"])
    return fx

def min_constraint_grad(x, sx):
    """ x[NUM_OBJ-1:NUM_DES] >= 0.55 """

    dx, ds = {}, {}
    for key in x:
        dx[key] = 0.
    for key in sx:
        ds[key] = 0.
    for i in range(NUM_OBJ - 1, NUM_DES):
        dx[f"x{i+1}"] = -1.0
    return dx, ds

def max_constraint_func(x, sx):
    """ x[NUM_OBJ-1:NUM_DES] <= 0.65 """

    fx = 0.0
    for i in range(NUM_OBJ - 1, NUM_DES):
        fx += (x[f"x{i+1}"] - 0.65)
    return fx

def max_constraint_grad(x, sx):
    """ x[NUM_OBJ-1:NUM_DES] <= 0.65 """

    dx, ds = {}, {}
    for key in x:
        dx[key] = 0.
    for key in sx:
        ds[key] = 0.
    for i in range(NUM_OBJ - 1, NUM_DES):
        dx[f"x{i+1}"] = 1.0
    return dx, ds

# Add 2 constraints to the problem
moop.addConstraint({'name': "Lower Bounds",
                    'con_func': min_constraint_func,
                    'con_grad': min_constraint_grad})
moop.addConstraint({'name': "Upper Bounds",
                    'con_func': max_constraint_func,
                    'con_grad': max_constraint_grad})

# Add 5 acquisition functions
for i in range(5):
    moop.addAcquisition({'acquisition': RandomConstraint, 'hyperparams': {}})

# Solve the problem with 5 iterations
moop.solve(5)

# Check that 125 simulations were evaluated and solutions are feasible
assert(moop.getObjectiveData().shape[0] == 125)
assert(moop.getSimulationData()['DTLZ1'].shape[0] == 125)
assert(moop.getPF().shape[0] > 0)



def test_static_export():
    """ Create a MOOP object and export various plot types.

    Tests all export formats except eps
    (because eps may not be supported on all machines).

    """

    from parmoo.viz.plot import (
        scatter,
        parallel_coordinates,
        radar,
    )
    import os

    # Pre-calculate a moop object
    moop1 = run_quickstart()

    # * html output
    scatter(moop1, output='html')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    # * svg output
    parallel_coordinates(moop1, output='svg')
    assert (os.path.exists("Pareto Front.svg"))
    os.remove("Pareto Front.svg")

    # * pdf output
    radar(moop1, output='pdf')
    assert (os.path.exists("Pareto Front.pdf"))
    os.remove("Pareto Front.pdf")

    # * jpeg output
    scatter(moop1, output='jpeg')
    assert (os.path.exists("Pareto Front.jpeg"))
    os.remove("Pareto Front.jpeg")

    # * png output
    parallel_coordinates(moop1, output='png')
    assert (os.path.exists("Pareto Front.png"))
    os.remove("Pareto Front.png")

    # * webp output
    radar(moop1, output='webp')
    assert (os.path.exists("Pareto Front.webp"))
    os.remove("Pareto Front.webp")


def test_quantity_constraints_objectives():
    """ Create a MOOP object and plot PF with and w/o constraint data. """

    from parmoo.viz.plot import (
        scatter,
        parallel_coordinates,
        radar,
    )
    import os

    # Pre-calculate two moop objects
    moop1 = run_quickstart()
    moop2 = run_dtlz2()

    # * 2 objective scatter with constraint
    scatter(moop1, output='html')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    # * 2 objective parallel with constraint
    parallel_coordinates(moop1, output='html')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    # * 2 objective radar with constraint
    radar(moop1, output='html')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    # * 5 objective scatter without constraint
    scatter(moop2, output='html')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    # * 5 objective parallel without constraint
    parallel_coordinates(moop2, output='html')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    # * 5 objective radar without constraint
    radar(moop2, output='html')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")


def test_database_options():
    """ Create a MOOP object and plot the full database.

    Generates plots with and without constraint violations.

    """

    from parmoo.viz.plot import (
        scatter,
        parallel_coordinates,
        radar,
    )
    import os

    # Pre-calculate two moop objects
    moop1 = run_quickstart()
    moop2 = run_dtlz2()

    # * pf x constraint_satisfying x constraints in MOOP
    scatter(
        moop1,
        db='pf',
        points='constraint_satisfying',
        output='html')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    parallel_coordinates(
        moop1,
        db='pf',
        points='constraint_satisfying',
        output='html')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    radar(
        moop1,
        db='pf',
        points='constraint_satisfying',
        output='html')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    # * obj x constraint_satisfying x constraints in MOOP
    scatter(
        moop1,
        db='obj',
        points='constraint_satisfying',
        output='html')
    assert (os.path.exists("Objective Data.html"))
    os.remove("Objective Data.html")

    parallel_coordinates(
        moop1,
        db='obj',
        points='constraint_satisfying',
        output='html')
    assert (os.path.exists("Objective Data.html"))
    os.remove("Objective Data.html")

    radar(
        moop1,
        db='obj',
        points='constraint_satisfying',
        output='html')
    assert (os.path.exists("Objective Data.html"))
    os.remove("Objective Data.html")

    # * pf x constraint_violating x constraints in MOOP
    scatter(
        moop1,
        db='pf',
        points='constraint_violating',
        output='html')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    parallel_coordinates(
        moop1,
        db='pf',
        points='constraint_violating',
        output='html')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    radar(
        moop1,
        db='pf',
        points='constraint_violating',
        output='html')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    # * obj x constraint_violating x constraints in MOOP
    scatter(
        moop1,
        db='obj',
        points='constraint_violating',
        output='html')
    assert (os.path.exists("Objective Data.html"))
    os.remove("Objective Data.html")

    parallel_coordinates(
        moop1,
        db='obj',
        points='constraint_violating',
        output='html')
    assert (os.path.exists("Objective Data.html"))
    os.remove("Objective Data.html")

    radar(
        moop1,
        db='obj',
        points='constraint_violating',
        output='html')
    assert (os.path.exists("Objective Data.html"))
    os.remove("Objective Data.html")

    # * pf x all x constraints in MOOP
    scatter(
        moop1,
        db='pf',
        points='all',
        output='html')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    parallel_coordinates(
        moop1,
        db='pf',
        points='all',
        output='html')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    radar(
        moop1,
        db='pf',
        points='all',
        output='html')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    # * obj x all x constraints in MOOP
    scatter(
        moop1,
        db='obj',
        points='all',
        output='html')
    assert (os.path.exists("Objective Data.html"))
    os.remove("Objective Data.html")

    parallel_coordinates(
        moop1,
        db='obj',
        points='all',
        output='html')
    assert (os.path.exists("Objective Data.html"))
    os.remove("Objective Data.html")

    radar(
        moop1,
        db='obj',
        points='all',
        output='html')
    assert (os.path.exists("Objective Data.html"))
    os.remove("Objective Data.html")

    # * pf x none x constraints in MOOP
    scatter(
        moop1,
        db='pf',
        points='none',
        output='html')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    parallel_coordinates(
        moop1,
        db='pf',
        points='none',
        output='html')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    radar(
        moop1,
        db='pf',
        points='none',
        output='html')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    # * obj x none x constraints in MOOP
    scatter(
        moop1,
        db='obj',
        points='none',
        output='html')
    assert (os.path.exists("Objective Data.html"))
    os.remove("Objective Data.html")

    parallel_coordinates(
        moop1,
        db='obj',
        points='none',
        output='html')
    assert (os.path.exists("Objective Data.html"))
    os.remove("Objective Data.html")

    radar(
        moop1,
        db='obj',
        points='none',
        output='html')
    assert (os.path.exists("Objective Data.html"))
    os.remove("Objective Data.html")

    # * pf x constraint_satisfying x no constraints in MOOP
    scatter(
        moop2,
        db='pf',
        points='constraint_satisfying',
        output='html')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    parallel_coordinates(
        moop2,
        db='pf',
        points='constraint_satisfying',
        output='html')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    radar(
        moop2,
        db='pf',
        points='constraint_satisfying',
        output='html')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    # * obj x constraint_satisfying x no constraints in MOOP
    scatter(
        moop2,
        db='obj',
        points='constraint_satisfying',
        output='html')
    assert (os.path.exists("Objective Data.html"))
    os.remove("Objective Data.html")

    parallel_coordinates(
        moop2,
        db='obj',
        points='constraint_satisfying',
        output='html')
    assert (os.path.exists("Objective Data.html"))
    os.remove("Objective Data.html")

    radar(
        moop2,
        db='obj',
        points='constraint_satisfying',
        output='html')
    assert (os.path.exists("Objective Data.html"))
    os.remove("Objective Data.html")

    # * pf x constraint_violating x no constraints in MOOP
    scatter(
        moop2,
        db='pf',
        points='constraint_violating',
        output='html')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    parallel_coordinates(
        moop2,
        db='pf',
        points='constraint_violating',
        output='html')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    radar(
        moop2,
        db='pf',
        points='constraint_violating',
        output='html')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    # * obj x constraint_violating x no constraints in MOOP
    scatter(
        moop2,
        db='obj',
        points='constraint_violating',
        output='html')
    assert (os.path.exists("Objective Data.html"))
    os.remove("Objective Data.html")

    parallel_coordinates(
        moop2,
        db='obj',
        points='constraint_violating',
        output='html')
    assert (os.path.exists("Objective Data.html"))
    os.remove("Objective Data.html")

    radar(
        moop2,
        db='obj',
        points='constraint_violating',
        output='html')
    assert (os.path.exists("Objective Data.html"))
    os.remove("Objective Data.html")

    # * pf x all x no constraints in MOOP
    scatter(
        moop2,
        db='pf',
        points='all',
        output='html')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    parallel_coordinates(
        moop2,
        db='pf',
        points='all',
        output='html')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    radar(
        moop2,
        db='pf',
        points='all',
        output='html')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    # * obj x all x no constraints in MOOP
    scatter(
        moop2,
        db='obj',
        points='all',
        output='html')
    assert (os.path.exists("Objective Data.html"))
    os.remove("Objective Data.html")

    parallel_coordinates(
        moop2,
        db='obj',
        points='all',
        output='html')
    assert (os.path.exists("Objective Data.html"))
    os.remove("Objective Data.html")

    radar(
        moop2,
        db='obj',
        points='all',
        output='html')
    assert (os.path.exists("Objective Data.html"))
    os.remove("Objective Data.html")

    # * pf x none x no constraints in MOOP
    scatter(
        moop2,
        db='pf',
        points='none',
        output='html')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    parallel_coordinates(
        moop2,
        db='pf',
        points='none',
        output='html')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    radar(
        moop2,
        db='pf',
        points='none',
        output='html')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    # * obj x none x no constraints in MOOP
    scatter(
        moop2,
        db='obj',
        points='none',
        output='html')
    assert (os.path.exists("Objective Data.html"))
    os.remove("Objective Data.html")

    parallel_coordinates(
        moop2,
        db='obj',
        points='none',
        output='html')
    assert (os.path.exists("Objective Data.html"))
    os.remove("Objective Data.html")

    radar(
        moop2,
        db='obj',
        points='none',
        output='html')
    assert (os.path.exists("Objective Data.html"))
    os.remove("Objective Data.html")


def test_inputs_to_dash():
    """ Stress-test the dash app's error handling. """

    from parmoo.viz.plot import scatter
    import os
    import pytest

    # Pre-calculate a moop object
    moop1 = run_quickstart()

    # * db
    # valid db values tested in test_database_options()
    # test invalid db values
    with pytest.raises(ValueError):
        scatter(moop1, output='html', db='asdf')

    with pytest.raises(ValueError):
        scatter(moop1, output='html', db=1234)

    # * output
    # valid output values tested in test_static_export()
    # test invalid output values
    with pytest.raises(ValueError):
        scatter(moop1, output='asdf')

    with pytest.raises(ValueError):
        scatter(moop1, output=1234)

    # * points
    # valid points values tested in test_database_options()
    # test invalid points values
    with pytest.raises(ValueError):
        scatter(moop1, output='html', points='asdf')

    with pytest.raises(ValueError):
        scatter(moop1, output='html', points=1234)

    # * height
    # test valid height values
    scatter(moop1, output='html', height=1)
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    scatter(moop1, output='html', height=50)
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    scatter(moop1, output='html', height=92.2678)
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    scatter(moop1, output='html', height=7789)
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    # test invalid height values
    with pytest.raises(ValueError):
        scatter(moop1, output='html', height='asdf')

    # * width
    # test valid width values
    scatter(moop1, output='html', width=1)
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    scatter(moop1, output='html', width=50)
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    scatter(moop1, output='html', width=92.2678)
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    scatter(moop1, output='html', width=7789)
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    # test invalid width values
    with pytest.raises(ValueError):
        scatter(moop1, output='html', width='asdf')

    # * font
    # test valid font values
    scatter(moop1, output='html', font='Verdana')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    scatter(moop1, output='html', font='Times New Roman')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    # test invalid font values
    # currently all inputs that can be cast to a string are valid
    # if str(font) != a font in the computer, font=Times New Roman

    # * fontsize
    # test valid fontsize values
    scatter(moop1, output='html', fontsize=1)
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    scatter(moop1, output='html', fontsize=50)
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    scatter(moop1, output='html', fontsize=55.71)
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    scatter(moop1, output='html', fontsize=100)
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    # test invalid fontsize values
    with pytest.raises(ValueError):
        scatter(moop1, output='html', fontsize=-1)

    with pytest.raises(ValueError):
        scatter(moop1, output='html', fontsize=101)

    # * background_color
    # test valid background_color values
    scatter(moop1, output='html', background_color='white')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    scatter(moop1, output='html', background_color='black')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    scatter(
        moop1,
        output='html',
        background_color='transparent'
    )
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    scatter(moop1, output='html', background_color='white')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    scatter(moop1, output='html', background_color='grey')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    # test invalid background_color values
    # currently all inputs that can be cast to a string are valid
    # if str(background_color) != a supported background_color
    # background_color=white

    # * screenshot
    # test valid screenshot values
    scatter(moop1, output='html', screenshot='png')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    scatter(moop1, output='html', screenshot='jpeg')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    scatter(moop1, output='html', screenshot='svg')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    scatter(moop1, output='html', screenshot='webp')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    # test invalid screenshot values
    with pytest.raises(ValueError):
        scatter(moop1, output='html', screenshot='asdf')

    with pytest.raises(ValueError):
        scatter(moop1, output='html', screenshot=1234)

    # * image_export_format
    # test valid image_export_format values (except eps)
    scatter(moop1, output='html', image_export_format='html')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    scatter(moop1, output='html', image_export_format='svg')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    scatter(moop1, output='html', image_export_format='pdf')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    scatter(moop1, output='html', image_export_format='png')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    scatter(moop1, output='html', image_export_format='jpeg')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    scatter(moop1, output='html', image_export_format='webp')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    # test invalid image_export_format values
    with pytest.raises(ValueError):
        scatter(moop1, output='html', image_export_format='asdf')

    with pytest.raises(ValueError):
        scatter(moop1, output='html', image_export_format=1234)

    # * data_export_format
    # test valid data_export_format values
    scatter(moop1, output='html', data_export_format='csv')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    scatter(moop1, output='html', data_export_format='json')
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    # test invalid data_export_format values
    with pytest.raises(ValueError):
        scatter(moop1, output='html', data_export_format='asdf')

    with pytest.raises(ValueError):
        scatter(moop1, output='html', data_export_format=1234)

    # * dev_mode
    # test valid dev_mode values
    scatter(moop1, output='html', dev_mode=True)
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    scatter(moop1, output='html', dev_mode=False)
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    # test invalid dev_mode values
    # currently all inputs that can be cast to a Boolean are valid
    # in Python, everything can be cast to a Boolean

    # * pop_up
    # test valid pop_up values
    scatter(moop1, output='html', pop_up=True)
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    scatter(moop1, output='html', pop_up=False)
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    # test invalid pop_up values
    # currently all inputs that can be cast to a Boolean are valid
    # in Python, everything can be cast to a Boolean

    # * port
    # test valid port values
    scatter(
        moop1,
        output='html',
        port='http://127.0.0.1:8050/'
    )
    assert (os.path.exists("Pareto Front.html"))
    os.remove("Pareto Front.html")

    # test invalid port values
    with pytest.raises(ValueError):
        scatter(moop1, output='html', port='asdf')

    with pytest.raises(ValueError):
        scatter(moop1, output='html', port=1234)


def run_quickstart():
    """ Auxiliary function that creates a MOOP by running the quickstart. """

    import numpy as np
    from parmoo import MOOP
    from parmoo.searches import LatinHypercube
    from parmoo.surrogates import GaussRBF
    from parmoo.acquisitions import UniformWeights
    from parmoo.optimizers import GlobalSurrogate_PS

    my_moop = MOOP(GlobalSurrogate_PS)

    my_moop.addDesign({
        'name': "x1",
        'des_type': "continuous",
        'lb': 0.0, 'ub': 1.0
    })
    my_moop.addDesign({
        'name': "x2",
        'des_type': "categorical",
        'levels': 3
    })

    def sim_func(x):
        if x["x2"] == 0:
            return np.array([(x["x1"] - 0.2) ** 2, (x["x1"] - 0.8) ** 2])
        else:
            return np.array([99.9, 99.9])

    my_moop.addSimulation({
        'name': "MySim",
        'm': 2,
        'sim_func': sim_func,
        'search': LatinHypercube,
        'surrogate': GaussRBF,
        'hyperparams': {'search_budget': 10}
    })

    my_moop.addObjective({
        'name': "Cost of driving",
        'obj_func': lambda x, s: s["MySim"][0]
    })
    my_moop.addObjective({
        'name': "Time spent driving",
        'obj_func': lambda x, s: s["MySim"][1]
    })

    my_moop.addConstraint({
        'name': "c1",
        'constraint': lambda x, s: 0.1 - x["x1"]
    })

    for i in range(2):
        my_moop.addAcquisition({
            'acquisition': UniformWeights,
            'hyperparams': {}
        })

    my_moop.solve(1)

    return my_moop


def run_dtlz2():
    """ Auxiliary function that creates a MOOP by running DTLZ2. """

    from parmoo import MOOP
    from parmoo.acquisitions import RandomConstraint
    from parmoo.searches import LatinHypercube
    from parmoo.surrogates import GaussRBF
    from parmoo.optimizers import GlobalSurrogate_BFGS
    from parmoo.objectives.dtlz import dtlz2_obj, dtlz2_grad
    from parmoo.simulations.dtlz import g2_sim

    n = 6  # number of design variables
    o = 5  # number of objectives
    q = 2  # batch size (number of acquisitions)
    # Create MOOP
    moop = MOOP(GlobalSurrogate_BFGS)
    # Add n design variables
    for i in range(n):
        moop.addDesign({
            'name': f"Input {i+1}",
            'des_type': 'continuous',
            'lb': 0.0,
            'ub': 1.0,
            'des_tol': 1.0e-8
        })
    # Create the g2 simulation
    moop.addSimulation({
        'name': "g2",
        'm': 1,
        'sim_func': g2_sim(
            moop.getDesignType(),
            num_obj=o,
            offset=0.5
        ),
        'search': LatinHypercube,
        'surrogate': GaussRBF,
        'hyperparams': {'search_budget': 10}
    })
    # Add o objectives
    for i in range(o):
        moop.addObjective({
            'name': f"Objective {i+1}",
            'obj_func': dtlz2_obj(
                moop.getDesignType(),
                moop.getSimulationType(),
                i, num_obj=o),
            'obj_grad': dtlz2_grad(
                moop.getDesignType(),
                moop.getSimulationType(),
                i, num_obj=o)
        })
    # Add q acquisition functions
    for i in range(q):
        moop.addAcquisition({'acquisition': RandomConstraint})
    # Solve the MOOP with 20 iterations
    moop.solve(1)
    return moop


if __name__ == "__main__":
    test_static_export()
    test_quantity_constraints_objectives()
    test_database_options()
    test_inputs_to_dash()
